#
# pyHbr.appliance.py
#
# General functions useful for working with OVF and
# deploying the HMS appliance.
#
import atexit
import logging
import os.path
import re
import signal
import socket
import struct
import subprocess
import sys

try:
   import simplejson as json
except ImportError:
   import json

import paramiko

# figure out the architecture using the pointer size
bits = str(struct.calcsize('P') * 8)
python_ver = 'python%d.%d' % (sys.version_info[0], sys.version_info[1])

sys.path.append('/build/apps/lib')
from build.utils import buildapi

logger = logging.getLogger('pyHbr.appliance')

def GetBuild(product=None, buildID=None, branch=None, kind='ob'):
   """Returns the URL to the AddOn OVF of the requested HMS build.

   If buildID was specified, the build information is fetched and
   the OVF URL is retrieved from it.
   If buildID is none, the build information for the latest saved
   build from the branch requested is fetched and the OVF URL is
   retrieved from it.

   @param buildID [in] the build identifier to get the OVF url for.
   @param branch  [in] branch to pickup build from.
   @param kind    [in] ob/sb (official/sandbox build)
   """
   if buildID:
      match = re.match('([os]b)-([0-9]+)', buildID)
      assert match is not None, 'Invalid build identifier: %s' % buildID

      kind = match.group(1)
      buildID = match.group(2)
      logger.debug('buildID is %s, kind is %s' % (buildID, kind))

   build = None
   api = buildapi.BuildApi(kind)
   if not buildID:
      query = api.Build.objects.all()
      query = query.filter(product=product, branch=branch, ondisk=1)

      build = query.order_by('-id')[0]
      if not build:
         logger.error('Cannot find a saved build for '\
                       'the product %s on branch %s' % (product, branch))
         return None
   else:
      build = api.Build.objects.get(pk=buildID)
      if not build.ondisk:
         logger.error('Build %s-%s is no longed '\
                       'saved on disk' % (kind, buildID))
         return None

   return build

def GetOvfURL(buildID=None, branch='vr-2016', kind='ob'):
   build = GetBuild('hms-va', buildID, branch, kind)
   return build._buildtree_url + '/publish/ovf/vSphere_Replication_AddOn_OVF10.ovf'


class OVFDeployment(object):
   """Base OVF deployment for the HMS virtual appliance.

   Abstract class that provides a base for an OVF deployment
   implementation of the HMS virtual appliance.
   """

   def __init__(self,
                filename=None,
                defaults=True):
      assert filename, "No OVF filename provided"

      if defaults:
         self.properties = {
               'enable_sshd': True,
               'password': 'vmware11'
               }
      else:
         self.properties = {}

      self.SetOVFFile(filename)


   def SetProperty(self, key, value):
      """Set an OVF property."""
      self.properties[key] = value

   def SetIpAllocationPolicy(self, ipAllocationPolicy):
      """Set the IP allocation policy."""
      raise NotImplementedError

   def SetIpProtocol(self, ipProtocol):
      """Set the IP protocol."""
      raise NotImplementedError

   def SetNetworkInfo(self, ip, netmask, gateway, dns):
      """Set the network details.

      Implicitly sets the IP allocation policy and protocol.
      """
      self.SetIpAllocationPolicy('fixedPolicy')

      # use inet_pton to check the type of the IP address
      try:
         socket.inet_pton(socket.AF_INET6, ip)
         self.SetIpProtocol('IPv6')
      except socket.error as e:
         socket.inet_pton(socket.AF_INET, ip)
         self.SetIpProtocol('IPv4')

      self.SetProperty('vami.ip0.vSphere_Replication_Appliance', ip)
      self.SetProperty('vami.netmask0.vSphere_Replication_Appliance', netmask)
      self.SetProperty('vami.gateway.vSphere_Replication_Appliance', gateway)
      self.SetProperty('vami.DNS.vSphere_Replication_Appliance', ','.join(dns))

   def SetSSHPassword(self, password):
      """Set the default SSH password for the appliance."""
      self.SetProperty('password', password)

   def SetDiskMode(self, diskMode):
      """Set the disk mode."""
      raise NotImplementedError

   def SetNetwork(self, network):
      """Set the network the VM nic will be attached to."""
      raise NotImplementedError

   def SetDatastore(self, datastore):
      """Set the datastore the VM will be deployed to."""
      raise NotImplementedError

   def SetOVFFile(self, filename):
      """Set the OVF file used for deployment"""
      self.filename = filename

   def SetName(self, name):
      """Set the name of the VM to be deployed"""
      self.SetOption('name', name)

   def Deploy(self):
      """Start OVF deployment.

      Returns a tuple that represents whether the deployment
      succeeded and the IP of the deployed VM.
      The second element of the tuple may be None if the
      IP could not be retrieved.
      """
      raise NotImplementedError


class OVFToolDeployment(OVFDeployment):
   """OVF deployment of the HMS virtual appliance using ovftool"""

   def __init__(self,
                filename=None,
                network=None,
                datastore=None,
                username=None,
                password=None,
                defaults=True,
                ovftoolBinary='/build/toolchain/lin%s/ovftool-4.0.0/ovftool' % bits,
                targetHost=None):
      OVFDeployment.__init__(self, filename, defaults)

      assert os.path.abspath(ovftoolBinary), "Path must be absolute: %s" % ovftoolBinary
      assert os.path.exists(ovftoolBinary), "Path does not exist: %s" % ovftoolBinary

      assert network, "No network provided"
      assert datastore, "No datastore provided"
      assert targetHost, "No target provided"
      assert username, "No username provided"

      if defaults:
         self.options = {
               'skipManifestCheck': None,
               'noSSLVerify': None,
               'acceptAllEulas': None,
               'powerOn': None,
               'X:waitForIp': None,
               'X:enableHiddenProperties': None,
               'diskMode': 'thick',
               'ipAllocationPolicy': 'dhcpPolicy',
               'ipProtocol': 'IPv4',
               }
      else:
         self.options = {}

      self.ovftoolBinary = ovftoolBinary
      self.SetTarget(targetHost, '/' in targetHost)
      self.SetUsername(username)
      self.SetPassword(password)
      self.SetNetwork(network)
      self.SetDatastore(datastore)

   def SetOption(self, key, value):
      """Set an option that will be passed to the ovftool"""
      self.options[key] = value

   def SetIpAllocationPolicy(self, ipAllocationPolicy):
      self.SetOption('ipAllocationPolicy', ipAllocationPolicy)

   def SetIpProtocol(self, ipProtocol):
      self.SetOption('ipProtocol', ipProtocol)

   def SetDiskMode(self, diskMode):
      self.SetOption('diskMode', diskMode)

   def SetNetwork(self, network):
      self.SetOption('network', network)

   def SetDatastore(self, datastore):
      self.SetOption('datastore', datastore)

   def SetUsername(self, username):
      """Set the username used for the target login."""
      self.username = username

   def SetPassword(self, password):
      """Set the password used for the target login."""
      self.password = password

   def SetTarget(self, targetHost, isVC=False):
      """Set the target for this deployment."""
      if not isVC and self.options['ipAllocationPolicy'] != 'fixedPolicy':
         self.SetOption('X:injectOvfEnv', None)

         #
         # This is a hack to prevent ovftool from sending VAMI some
         # bogus values for these properties when deploying directly to
         # the ESX host. The problem is that ovftool expects these values
         # to be further replaced by the VC so if they are not set, it
         # sends some placeholders which the VAMI cannot handle.
         #
         self.SetProperty('vami.ip0.vSphere_Replication_Appliance', '')
         self.SetProperty('vami.netmask0.vSphere_Replication_Appliance', '')
         self.SetProperty('vami.gateway.vSphere_Replication_Appliance', '')
         self.SetProperty('vami.DNS.vSphere_Replication_Appliance', '')

      self.targetHost = targetHost

   def GetTarget(self):
      """Get the target used for this deployment"""
      assert self.targetHost, "Need target"
      assert self.username, "Need username"

      return 'vi://%s:%s@%s' % (self.username, self.password, self.targetHost)

   def Deploy(self):
      command = [self.ovftoolBinary]

      options = []
      for key, value in self.options.iteritems():
         if value:
            options += ['--%s=%s' % (key, value)]
         else:
            options += ['--%s' % key]

      for key, value in self.properties.iteritems():
         if value:
            options += ['--prop:%s=%s' % (key, value)]
         else:
            options += ['--prop:%s' % key]

      command += options + [self.filename, self.GetTarget()]

      ip = None
      pattern = re.compile('Received IP address: ([0-9.]*)')

      logger.debug('Running command %s' % " ".join(command))
      process = subprocess.Popen(command, bufsize=0,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.STDOUT)
      def kill_ovftool():
         try:
            os.kill(process.pid, signal.SIGTERM)
         except OSError:
            pass
      atexit.register(kill_ovftool)

      #
      # Print all the output from the ovftool to provide
      # progress for the user and check every line for
      # the IP assigned to the VM
      #
      content = ''
      while True:
         buffer = process.stdout.read(1)
         if len(buffer) == 0:
            break

         sys.stdout.write(buffer)
         sys.stdout.flush()
         content += buffer

      match = pattern.search(content)
      if match is not None:
         ip = match.group(1)
         logger.debug('Found IP "%s"' % ip)

      if process.poll() != 0:
         logger.error('Failed to run ovftool ovftool command')
         return (False, None)

      if not ip:
         logger.warning('Failed to get IP address of the deployment')

      return (True, ip)


