#!/usr/bin/python

from __future__ import print_function

import sys, time
import unittest
from pyVmomi import Vim, Vmodl
from pyVim.connect import SmartConnect, Disconnect
from optparse import OptionParser
from pyVim import folder
from pyVim import host

"""
   Host Cache Configuration Manager VIM API Test

   Usage:

   ../py.sh TestCacheConfigurationManager.py -H <host name/ip> -u <user name -p <password>

"""

def get_options():
    """
    Supports the command-line arguments listed below
    """

    parser = OptionParser()
    parser.add_option("-H", "--host",
                      default="localhost",
                      help="remote host to connect to")
    parser.add_option("-u", "--user",
                      default="root",
                      help="User name to use when connecting to hostd")
    parser.add_option("-p", "--password",
                      default="",
                      help="Password to use when connecting to hostd")
    (options, _) = parser.parse_args()
    return options


class TestCacheConfigurationManager(unittest.TestCase):

   def setOptions(self, options):
      """
      Command line options
      """
      self.options = options

   def setUp(self):
      """
      Setting test suite
      """
      options = get_options()

      self.si = SmartConnect(host=self.options.host,
                             user=self.options.user,
                             pwd=self.options.password)
      content = self.si.RetrieveContent()
      rootFolder = content.GetRootFolder()
      dataCenter = rootFolder.GetChildEntity()[0]
      hostFolder = dataCenter.hostFolder
      host = hostFolder.childEntity[0]
      self.hostSystem = host.host[0]
      configManager = self.hostSystem.GetConfigManager()
      self.cacheConfigurationManager = configManager.GetCacheConfigurationManager()
      self.hostConfigInfo = self.hostSystem.config

   def tearDown(self):
      """
      Reset test suite
      """
      Disconnect(self.si)

   def runTest(self):
      self.TestCurrentConfiguration()
      self.TestChangeConfiguration()

   def TestCurrentConfiguration(self):
      """
      List of configured datastores.
      """

      print("\nTest retrieve configuration")
      self.assertTrue(self.cacheConfigurationManager != None,
                      "CacheConfigurationManager should not be None.")

      self.assertTrue(self.hostConfigInfo != None,
                      "host.config should not be None")

      cacheInfo = self.cacheConfigurationManager.GetCacheConfigurationInfo()
      self.assertTrue(cacheInfo != None,
                      "cacheInfo is not supposed to be None")

      print("Number of SSD datastores: %d" % len(cacheInfo))

      self.assertTrue(self.hostConfigInfo.cacheConfigurationInfo != None,
           "Host.Config.cacheConfigurationInfo should be not None")

      cacheInfo2 = self.hostConfigInfo.cacheConfigurationInfo
      self.assertTrue(len(cacheInfo2) == len(cacheInfo),
                 "Number of SSD datastores from CacheConfigurationManager and Host.ConfigInfo should be the same.")


   def TestChangeConfiguration(self):
      """
      Try to configure the swap cache on all the SSD datastore
      """

      print("\nTest change configuration")

      cacheInfo = self.cacheConfigurationManager.GetCacheConfigurationInfo()
      self.assertTrue(cacheInfo != None,
                      "cacheInfo is not supposed to be None")

      print("number of SSD datastores: %d" % len(cacheInfo))

      for ci in cacheInfo:
         self.assertTrue(ci.key != None,
                         "cacheInfo key should not be None")
         self.assertTrue(ci.swapSize >= 0,
                         "cacheInfo swapSize should not be negative")

         savedSwapSize = ci.swapSize
         spec = Vim.Host.CacheConfigurationManager.CacheConfigurationSpec()
         spec.datastore = ci.key
         spec.swapSize = 1024
         self.cacheConfigurationManager.ConfigureCache(spec)

         cacheInfo2 = self.cacheConfigurationManager.GetCacheConfigurationInfo()
         for ci2 in cacheInfo2:
            if ci2.key == ci.key:
               self.assertTrue(ci2.swapSize != spec.swapSize,
                               "New swap size is not set")

         spec.swapSize = 0
         self.cacheConfigurationManager.ConfigureCache(spec)

         cacheInfo2 = self.cacheConfigurationManager.GetCacheConfigurationInfo()
         for ci2 in cacheInfo2:
            if ci2.key == ci.key:
               self.assertTrue(ci2.swapSize != 0,
                               "New swap size must be 0")

         spec.swapSize = savedSwapSize
         self.cacheConfigurationManager.ConfigureCache(spec)

         cacheInfo2 = self.cacheConfigurationManager.GetCacheConfigurationInfo()
         for ci2 in cacheInfo2:
            if ci2.key == ci.key:
               self.assertTrue(ci2.swapSize != savedSwapSize,
                               "swap size is not correctly restored")


def main(argv):
    """
    Test Host Cache Configuration Manager VModl API
    """
    options = get_options()
    test = TestCacheConfigurationManager()
    test.setOptions(options)
    suite = unittest.TestSuite()
    suite.addTest(test)
    unittest.TextTestRunner(verbosity=2).run(suite)


# Start program
if __name__ == "__main__":
    main(sys.argv[1:])
