#!/usr/bin/python

from __future__ import print_function

import sys, time
import unittest
from pyVmomi import Vim, Vmodl
from pyVim.connect import SmartConnect, Disconnect
from optparse import OptionParser
from pyVim import folder
from pyVim import host

"""
   Drs Stats Configuration Manager VIM API Test

   Usage:

   ../py.sh TestDrsStats.py -H <host name/ip> -u <user name -p <password>

"""

def get_options():
    """
    Supports the command-line arguments listed below
    """

    parser = OptionParser()
    parser.add_option("-H", "--host",
                      default="localhost",
                      help="remote host to connect to")
    parser.add_option("-u", "--user",
                      default="root",
                      help="User name to use when connecting to hostd")
    parser.add_option("-p", "--password",
                      default="",
                      help="Password to use when connecting to hostd")
    (options, _) = parser.parse_args()
    return options


class TestManager(unittest.TestCase):

   def setOptions(self, options):
      """
      Command line options
      """
      self.options = options

   def setUp(self):
      """
      Setting test suite
      """
      options = get_options()

      self.si = SmartConnect(host=self.options.host,
                             user=self.options.user,
                             pwd=self.options.password)
      self.internalContent = self.si.RetrieveInternalContent()

   def tearDown(self):
      """
      Reset test suite
      """
      Disconnect(self.si)

   def runTest(self):
      self.TestCurrentState()
      self.TestConfigureParams()

   def TestCurrentState(self):
      """
      General checks that all the property values are
      populated.
      """
      print("Testing current state ...")
      self.assertTrue(self.internalContent != None,
                      "Internal Service Content should not be none")
      drsStatsManager = self.internalContent.drsStatsManager

      self.assertTrue(drsStatsManager != None,
                      "DRS stats manager should not be none")

      params = drsStatsManager.QueryWorkloadCharacterization()
      for param in params:
         self.assertTrue(param.key != None,
                         "datastore key should not be none")
         self.assertTrue(param.outstandingIo > 0,
                         "outstandingIo should be > 0")
         self.assertTrue(param.ioSize > 0,
                         "ioSize should be > 0")
         self.assertTrue(param.readPercent == 0,
                         "read percent should be 0")
         self.assertTrue(param.randomPercent > 0,
                         "random percent should be > 0")

      print("Done.")

   def TestConfigureParams(self):
      """
      Try to reconfigure BASIL parameters
      """

      print("Testing configuration method ...")
      drsStatsManager = self.internalContent.drsStatsManager
      params = drsStatsManager.QueryWorkloadCharacterization()
      storedParams = []
      testParams = []

      for param in params:
         wParam = Vim.DrsStatsManager.WorkloadCharacterization()
         wParam.key = param.key
         wParam.outstandingIo = param.outstandingIo
         wParam.ioSize = param.ioSize
         wParam.readPercent = param.readPercent
         wParam.randomPercent = param.randomPercent
         storedParams.append(wParam)

         tParam = Vim.DrsStatsManager.WorkloadCharacterization()
         tParam.key = param.key
         tParam.outstandingIo = 1.
         tParam.ioSize = 2.
         tParam.readPercent = 3.
         tParam.randomPercent = 4.
         testParams.append(tParam)

      drsStatsManager.ConfigureWorkloadCharacterization(testParams)

      params = drsStatsManager.QueryWorkloadCharacterization()
      for param in params:
         self.assertTrue(param.key != None,
                         "datastore key should not be none")
         self.assertTrue((param.outstandingIo - 1.) < 0.001,
                         "outstanding i/o should be set to 1")
         self.assertTrue((param.ioSize - 2.) < 0.001,
                         "i/o size should be set to 2")
         self.assertTrue((param.readPercent - 3.) < 0.001,
                         "unused read percent should be set to 3")
         self.assertTrue((param.randomPercent - 4.) < 0.001,
                         "random percent should be set to 4")

      drsStatsManager.ConfigureWorkloadCharacterization(storedParams)

      print("Done.")

def main(argv):
    """
    Test DRS Stats Manager VModl API
    """
    options = get_options()
    test = TestManager()
    test.setOptions(options)
    suite = unittest.TestSuite()
    suite.addTest(test)
    unittest.TextTestRunner(verbosity=2).run(suite)


# Start program
if __name__ == "__main__":
    main(sys.argv[1:])
