#!/usr/bin/python

from __future__ import print_function

import sys
from pyVmomi import Vim, Vmodl
from pyVim.connect import Connect, Disconnect
from optparse import OptionParser
from pyVim.task import WaitForTask
from pyVim import folder
from pyVim import vm, host
from pyVim import invt
from pyVim import vmconfig
from pyVim import vimutil
from pyVim import arguments
from pyVim.helpers import StopWatch, Log
import time
import atexit

invt_path = "ha-datacenter/datastore/"

def get_options():
    """
    Supports the command-line arguments listed below
    """

    parser = OptionParser()
    parser.add_option("-H", "--host",
                      default="localhost",
                      help="remote host to connect to")
    parser.add_option("-u", "--user",
                      default="root",
                      help="User name to use when connecting to hostd")
    parser.add_option("-p", "--password",
                      default="",
                      help="Password to use when connecting to hostd")
    parser.add_option("-s", "--storage",
                      default="datastore1",
                      help="Storage to perform test on");
    parser.add_option("-v", "--vmName", default="VM1",
                      help="vm name")
    (options, _) = parser.parse_args()
    return options

def setup(si, vmName, dsName):
    """
    Clean up old and create new virtual machine
    with three disks
    """

    cleanup(si, vmName)

    vm.CreateQuickDummy(vmName, 3, diskSizeInMB = 128, datastoreName = dsName)
    searchIndex = si.RetrieveContent().searchIndex
    vm1 = searchIndex.FindByInventoryPath("ha-datacenter/vm/" + vmName)
    vm.PowerOn(vm1)



def cleanup(si, vmName):
   """
   Remove vm created.
   """

   searchIndex = si.RetrieveContent().searchIndex
   vm1 = searchIndex.FindByInventoryPath("ha-datacenter/vm/" + vmName)
   try:
      vm.PowerOff(vm1)
   except:
      pass
   if (vm1 != None):
      vm1.Destroy()


def TestDatastoreIORM1(si, expectedCongestionThreshold, expectedEnabled):
   """
   Test IORM setting on default datastore:
   a) Enable/disable IORM
   b) Set/Get congestion threshold

   datastore1 is an assumed name for vmfs datastore

   """

   global invt_path
   searchIndex = si.RetrieveContent().searchIndex
   datastore = searchIndex.FindByInventoryPath(invt_path)
   iormc = datastore.iormConfiguration
   if (iormc == None):
      raise Exception("iormConfiguration is None")

   if (not datastore.capability.storageIORMSupported):
      raise Exception("storage io rm is not supported on this datastore")

   spec = Vim.StorageResourceManager.IORMConfigSpec()
   if (datastore.iormConfiguration.enabled == expectedEnabled):
      expectedEnabled = not expectedEnabled
   spec.enabled = expectedEnabled

   if (datastore.iormConfiguration.congestionThreshold == expectedCongestionThreshold):
      expectedCongestionThreshold = expectedCongestionThreshold + 1000
   spec.congestionThreshold = expectedCongestionThreshold

   srm = si.RetrieveContent().storageResourceManager

   srm.ConfigureDatastoreIORMOnHost(datastore, spec)

   iormc = datastore.iormConfiguration
   if (iormc == None):
      raise Exception("iormConfiguration is None")

   if (expectedCongestionThreshold != iormc.congestionThreshold):
      raise Exception("congestionThreshold mismatch")

   if (expectedEnabled != iormc.enabled):
      raise Exception("enabled mismatch")

def TestDatastoreIORM2(si, expectedCongestionThreshold):
   """
   Test IORM setting on default datastore:
   a) Enable/disable IORM
   b) Set/Get congestion threshold

   datastore1 is an assumed name for vmfs datastore.  This test modifies
   only congestion threshold

   """

   global invt_path
   searchIndex = si.RetrieveContent().searchIndex
   datastore = searchIndex.FindByInventoryPath(invt_path)
   iormc = datastore.iormConfiguration
   if (iormc == None):
      raise Exception("iormConfiguration is None")

   if (not datastore.capability.storageIORMSupported):
      raise Exception("storage io rm is not supported on this datastore")

   spec = Vim.StorageResourceManager.IORMConfigSpec()
   expectedEnabled = datastore.iormConfiguration.enabled
   spec.enabled = None

   if (datastore.iormConfiguration.congestionThreshold == expectedCongestionThreshold):
      expectedCongestionThreshold = expectedCongestionThreshold + 1000
   spec.congestionThreshold = expectedCongestionThreshold

   srm = si.RetrieveContent().storageResourceManager

   srm.ConfigureDatastoreIORMOnHost(datastore, spec)

   iormc = datastore.iormConfiguration
   if (iormc == None):
      raise Exception("iormConfiguration is None")

   if (expectedCongestionThreshold != iormc.congestionThreshold):
      raise Exception("congestionThreshold mismatch")

   if (expectedEnabled != iormc.enabled):
      raise Exception("enabled mismatch")

def TestDatastoreIORM3(si, expectedEnabled):
   """
   Test IORM setting on default datastore:
   a) Enable/disable IORM
   b) Set/Get congestion threshold

   datastore1 is an assumed name for vmfs datastore.  This test only
   enables/disables IORM without changing congestion threshold.

   """

   global invt_path
   searchIndex = si.RetrieveContent().searchIndex
   datastore = searchIndex.FindByInventoryPath(invt_path)
   iormc = datastore.iormConfiguration
   if (iormc == None):
      raise Exception("iormConfiguration is None")

   if (not datastore.capability.storageIORMSupported):
      raise Exception("storage io rm is not supported on this datastore")

   spec = Vim.StorageResourceManager.IORMConfigSpec()
   if (datastore.iormConfiguration.enabled == expectedEnabled):
      expectedEnabled = not expectedEnabled
   spec.enabled = expectedEnabled

   expectedCongestionThreshold = datastore.iormConfiguration.congestionThreshold

   srm = si.RetrieveContent().storageResourceManager

   srm.ConfigureDatastoreIORMOnHost(datastore, spec)

   iormc = datastore.iormConfiguration
   if (iormc == None):
      raise Exception("iormConfiguration is None")

   if (expectedCongestionThreshold != iormc.congestionThreshold):
      raise Exception("congestionThreshold mismatch")

   if (expectedEnabled != iormc.enabled):
      raise Exception("enabled mismatch")



def TestVirtualDiskShares(si, vmName, newLevel, newLimit, newRes):
   """
   Check set/get IORM shares and limits on vm virtual disk
   """

   searchIndex = si.RetrieveContent().searchIndex
   vm = searchIndex.FindByInventoryPath("ha-datacenter/vm/" + vmName)
   if (vm == None):
       raise Exception("created vm is not found")

   disks = vmconfig.CheckDevice(vm.GetConfig(), Vim.Vm.Device.VirtualDisk)
   if len(disks) < 1:
      raise Exception("Failed to find a virtual disk")

   # Prepare expected results
   cspec = Vim.vm.ConfigSpec()
   expected = []
   print("Setting disk shares")
   print("Level %s"  % newLevel)
   print("Limit %s" % newLimit)
   print("Reservation %s" % newRes)
   for i in range(0, len(disks)):
      disk = disks[i]

      if (disk.storageIOAllocation == None):
         raise Exception("Uninitialized storageAllocation")

      if (disk.storageIOAllocation.shares.level == newLevel):
         newLevel = Vim.SharesInfo.Level.low
      disk.storageIOAllocation.shares.level = newLevel

      if (disk.storageIOAllocation.limit == newLimit):
         newLimit = 4000L
      disk.storageIOAllocation.limit = newLimit

      disk.storageIOAllocation.reservation = newRes

      expected.append((newLevel, newLimit, newRes))

      dspec = Vim.Vm.Device.VirtualDeviceSpec()
      dspec.device = disk
      dspec.operation = Vim.Vm.Device.VirtualDeviceSpec.Operation.edit
      cspec.deviceChange.append(dspec)

   task = vm.Reconfigure(cspec)
   WaitForTask(task)


   # Verify results
   disks = vmconfig.CheckDevice(vm.GetConfig(), Vim.Vm.Device.VirtualDisk)
   if len(disks) < 1:
      raise Exception("Failed to find a virtual disk")
   for i in range(0, len(disks)):
      disk = disks[i]

      if (disk.storageIOAllocation == None):
         raise Exception("Uninitialized storageAllocation")

      (expectedLevel, expectedLimit, expectedRes) = expected[i]
      if (disk.storageIOAllocation.shares.level != expectedLevel):
         raise Exception("Level does not match")
      print("Returned shares level: %s" % disk.storageIOAllocation.shares.level)

      if (disk.storageIOAllocation.limit != expectedLimit):
         raise Exception("Limit does not match")
      print("Returned limit: %s" % disk.storageIOAllocation.limit)

      if (disk.storageIOAllocation.reservation != expectedRes):
         raise Exception("Reservation does not match")
      print("Returned reservation: %s" % disk.storageIOAllocation.reservation)


def TestDeprecatedVirtualDiskShares(si, vmName, newLevel, newLimit):
   """
   Check set/get IORM shares in an old way using VirtualDisk.shares
   property
   """

   searchIndex = si.RetrieveContent().searchIndex
   vm = searchIndex.FindByInventoryPath("ha-datacenter/vm/" + vmName)
   if (vm == None):
       raise Exception("created vm is not found")

   disks = vmconfig.CheckDevice(vm.GetConfig(), Vim.Vm.Device.VirtualDisk)
   if len(disks) < 1:
      raise Exception("Failed to find a virtual disk")

   # Prepare expected results
   cspec = Vim.vm.ConfigSpec()
   expected = []
   for i in range(0, len(disks)):
      disk = disks[i]

      disk.storageIOAllocation = None

      if (disk.shares.level == newLevel):
         newLevel = Vim.SharesInfo.Level.low
      disk.shares.level = newLevel

      expected.append(newLevel)

      dspec = Vim.Vm.Device.VirtualDeviceSpec()
      dspec.device = disk
      dspec.operation = Vim.Vm.Device.VirtualDeviceSpec.Operation.edit
      cspec.deviceChange.append(dspec)

   task = vm.Reconfigure(cspec)
   WaitForTask(task)


   # Verify results
   disks = vmconfig.CheckDevice(vm.GetConfig(), Vim.Vm.Device.VirtualDisk)
   if len(disks) < 1:
      raise Exception("Failed to find a virtual disk")
   for disk in disks:

      if (disk.shares == None):
         raise Exception("Uninitialized VirtualDisk.shares")

      expectedLevel = expected[i]
      if (disk.shares.level != expectedLevel):
         raise Exception("Level does not match")


def TestOutOfLimitsVirtualDiskShares(si, vmName, wrongShares, correct):
   """
   Check that a proper exception is thrown when we pass out of
   the limit value (200-4000) for disk shares.
   """

   searchIndex = si.RetrieveContent().searchIndex
   vm = searchIndex.FindByInventoryPath("ha-datacenter/vm/" + vmName)
   if (vm == None):
       raise Exception("created vm is not found")

   disks = vmconfig.CheckDevice(vm.GetConfig(), Vim.Vm.Device.VirtualDisk)
   if len(disks) < 1:
      raise Exception("Failed to find a virtual disk")

   # Prepare configuration parameters
   cspec = Vim.vm.ConfigSpec()
   for i in range(0, len(disks)):
      disk = disks[i]
      disk.storageIOAllocation = None
      disk.shares.level = Vim.SharesInfo.Level.custom
      disk.shares.shares = wrongShares
      dspec = Vim.Vm.Device.VirtualDeviceSpec()
      dspec.device = disk
      dspec.operation = Vim.Vm.Device.VirtualDeviceSpec.Operation.edit
      cspec.deviceChange.append(dspec)

   error = None
   try:
      task = vm.Reconfigure(cspec)
      WaitForTask(task)
   except Exception as e:
      error = e

   if (correct):
      if (error != None):
         raise Exception("Exception generated for correct value(%d): %s" % (wrongShares, error))
   elif (not isinstance(error, Vmodl.Fault.InvalidArgument)):
      raise Exception("Wrong exception generated: %s" % error)
   else:
      print("Correct invalid argument exception: %s" % error)


def TestWrongCustomVirtualDiskShares(si, vmName):
   """
   Try to set wrong value for IORM shares on virtual disk
   """

   searchIndex = si.RetrieveContent().searchIndex
   vm = searchIndex.FindByInventoryPath("ha-datacenter/vm/" + vmName)
   if (vm == None):
       raise Exception("created vm is not found")

   disks = vmconfig.CheckDevice(vm.GetConfig(), Vim.Vm.Device.VirtualDisk)
   if len(disks) < 1:
      raise Exception("Failed to find a virtual disk")

   # Prepare expected results
   cspec = Vim.vm.ConfigSpec()
   expected = []
   for i in range(0, len(disks)):
      disk = disks[i]
      disk.storageIOAllocation = Vim.StorageResourceManager.IOAllocationInfo()
      disk.storageIOAllocation.shares = Vim.SharesInfo()
      disk.storageIOAllocation.shares.level = Vim.SharesInfo.Level.custom
      disk.storageIOAllocation.shares.shares = 199

      # Do not set limit - it should be marshalled as None
      # disk.storageIOAllocation.limit = 4000
      #

      dspec = Vim.Vm.Device.VirtualDeviceSpec()
      dspec.device = disk
      dspec.operation = Vim.Vm.Device.VirtualDeviceSpec.Operation.edit
      cspec.deviceChange.append(dspec)

   task = vm.Reconfigure(cspec)
   error = None
   try:
      WaitForTask(task)
   except Exception as e:
      error = e
   if (not isinstance(error, Vmodl.Fault.InvalidArgument)):
      raise Exception("Wrong exception generated: %s" % error)



def TestWrongCustomVirtualDiskLimit(si, vmName, wrongLimit, isCorrect):
   """
   Try to set wrong limit value
   """

   searchIndex = si.RetrieveContent().searchIndex
   vm = searchIndex.FindByInventoryPath("ha-datacenter/vm/" + vmName)
   if (vm == None):
       raise Exception("created vm is not found")

   disks = vmconfig.CheckDevice(vm.GetConfig(), Vim.Vm.Device.VirtualDisk)
   if len(disks) < 1:
      raise Exception("Failed to find a virtual disk")

   # Prepare expected results
   cspec = Vim.vm.ConfigSpec()
   expected = []
   for i in range(0, len(disks)):
      disk = disks[i]
      disk.storageIOAllocation = Vim.StorageResourceManager.IOAllocationInfo()

      # Do not set shares
      disk.storageIOAllocation.shares = None

      disk.storageIOAllocation.limit = wrongLimit

      dspec = Vim.Vm.Device.VirtualDeviceSpec()
      dspec.device = disk
      dspec.operation = Vim.Vm.Device.VirtualDeviceSpec.Operation.edit
      cspec.deviceChange.append(dspec)

   task = vm.Reconfigure(cspec)
   error = None
   try:
      WaitForTask(task)
   except Exception as e:
      error = e
   if (isCorrect and error != None):
      raise Exception("Exception %s thrown for value %d" % (error, wrongLimit))
   if (not isCorrect and not isinstance(error, Vmodl.Fault.InvalidArgument)):
      raise Exception("Wrong exception %s for value %d"  % (error, wrongLimit))



def TestConfigurationOptions(si, vmName):
   """
   Test configuration options which return value range and default
   values.
   """

   searchIndex = si.RetrieveContent().searchIndex
   content = si.RetrieveContent()
   rootFolder = content.GetRootFolder()
   dataCenter = rootFolder.GetChildEntity()[0]
   hostFolder = dataCenter.hostFolder
   host = hostFolder.childEntity[0]
   hostSystem = host.host[0]

   srm = si.RetrieveContent().storageResourceManager
   print(srm.QueryIORMConfigOption(hostSystem))

   vm = searchIndex.FindByInventoryPath("ha-datacenter/vm/" + vmName)
   if (vm == None):
       raise Exception("created vm is not found")

   option = vm.environmentBrowser.QueryConfigOption(None, None)

   if (option.hardwareOptions.resourceConfigOption == None):
      raise Exception("hardwareOptions.resourceConfigOption == None")
   rco = option.hardwareOptions.resourceConfigOption
   if (rco.cpuAllocationOption == None):
      raise Exception("cpuAllocationOption == None")
   if (rco.memoryAllocationOption == None):
      raise Exception("memoryAllocationOption == None")
   for vdo in option.hardwareOptions.virtualDeviceOption:
      if isinstance(vdo, Vim.Vm.Device.VirtualDiskOption):
         if (vdo.ioAllocationOption == None):
            raise Exception("ioAllocationOption is not initialized")

def TestIncorrectCongestionThreshold(si, wrongValue):
   """
   Negative test to check invalid parameter handling for congestion
   threshold.
   """

   global invt_path
   searchIndex = si.RetrieveContent().searchIndex
   datastore = searchIndex.FindByInventoryPath(invt_path)
   iormc = datastore.iormConfiguration
   if (iormc == None):
      raise Exception("iormConfiguration is None")

   if (not datastore.capability.storageIORMSupported):
      raise Exception("storage io rm is not supported on this datastore")

   spec = Vim.StorageResourceManager.IORMConfigSpec()
   spec.congestionThreshold = wrongValue
   srm = si.RetrieveContent().storageResourceManager
   error = None
   try:
      srm.ConfigureDatastoreIORMOnHost(datastore, spec)
   except Exception as e:
      error = e
   if (not isinstance(error, Vmodl.Fault.InvalidArgument)):
      raise Exception("Wrong exception %s for value %d"  % (error, wrongValue))

def TestRemoveDatastore(si):
   global invt_path
   searchIndex = si.RetrieveContent().searchIndex
   datastore = searchIndex.FindByInventoryPath(invt_path)
   iormc = datastore.iormConfiguration
   if (iormc == None):
      raise Exception("iormConfiguration is None")

   if (not datastore.capability.storageIORMSupported):
      raise Exception("storage io rm is not supported on this datastore")

   spec = Vim.StorageResourceManager.IORMConfigSpec()
   spec.enabled = True

   srm = si.RetrieveContent().storageResourceManager
   srm.ConfigureDatastoreIORMOnHost(datastore, spec)

   content = si.RetrieveContent()
   rootFolder = content.GetRootFolder()
   dataCenter = rootFolder.GetChildEntity()[0]
   hostFolder = dataCenter.hostFolder
   host = hostFolder.childEntity[0]
   hostSystem = host.host[0]
   configManager = hostSystem.GetConfigManager()
   datastoreSystem = configManager.datastoreSystem
   error = None
   try:
      task = datastoreSystem.RemoveDatastore(datastore)
      WaitForTask(task)
   except Exception as e:
      error = e
   if (not isinstance(error, Vim.Fault.ResourceInUse)):
      raise Exception("Wrong exception %s"  % error)


def TestDeprecatedRemoveDatastore(si):
   global invt_path
   searchIndex = si.RetrieveContent().searchIndex
   datastore = searchIndex.FindByInventoryPath(invt_path)
   iormc = datastore.iormConfiguration
   if (iormc == None):
      raise Exception("iormConfiguration is None")

   if (not datastore.capability.storageIORMSupported):
      raise Exception("storage io rm is not supported on this datastore")

   spec = Vim.StorageResourceManager.IORMConfigSpec()
   spec.enabled = True

   srm = si.RetrieveContent().storageResourceManager
   srm.ConfigureDatastoreIORMOnHost(datastore, spec)

   content = si.RetrieveContent()
   rootFolder = content.GetRootFolder()
   dataCenter = rootFolder.GetChildEntity()[0]
   hostFolder = dataCenter.hostFolder
   host = hostFolder.childEntity[0]
   hostSystem = host.host[0]
   configManager = hostSystem.GetConfigManager()
   datastoreSystem = configManager.datastoreSystem
   error = None
   try:
      datastore.DestroyDatastore()
   except Exception as e:
      error = e
   if (not isinstance(error, Vim.Fault.ResourceInUse)):
      raise Exception("Wrong exception %s"  % error)


def main():
    """
    Test IORM-related VModl API
    """
    options = get_options()
    si = Connect(host=options.host,
                 user=options.user,
                 version="vim.version.version9",
                 pwd=options.password)
    atexit.register(Disconnect, si)
    global invt_path
    storage = options.storage
    print("Using datastore : " + storage)
    invt_path = invt_path + storage
    print("Using path : " + invt_path)

    try:
       print("cleaning up")
       cleanup(si, options.vmName)
    except Exception:
       print("checked for vm from previous run")

    print("setting up vm")
    setup(si, options.vmName, storage)

    try:
        TestDatastoreIORM1(si, 15, True)
        TestDatastoreIORM1(si, 70, False)
        TestDatastoreIORM2(si, 15)
        TestDatastoreIORM2(si, 70)
        TestDatastoreIORM3(si, True)
        TestDatastoreIORM3(si, False)

        TestVirtualDiskShares(si, options.vmName, Vim.SharesInfo.Level.normal, 3000L, 1500)
        TestVirtualDiskShares(si, options.vmName, Vim.SharesInfo.Level.high, 5000L, 0)
        TestVirtualDiskShares(si, options.vmName, Vim.SharesInfo.Level.low, 2000L, 10)

        TestDeprecatedVirtualDiskShares(si, options.vmName, Vim.SharesInfo.Level.normal, 3000L)
        TestDeprecatedVirtualDiskShares(si, options.vmName, Vim.SharesInfo.Level.high, 5000L)
        TestDeprecatedVirtualDiskShares(si, options.vmName, Vim.SharesInfo.Level.low, 2000L)

        TestOutOfLimitsVirtualDiskShares(si, options.vmName, 199, False)
        TestOutOfLimitsVirtualDiskShares(si, options.vmName, 4001, False)
        TestOutOfLimitsVirtualDiskShares(si, options.vmName, 3000, True)
        TestOutOfLimitsVirtualDiskShares(si, options.vmName, 200, True)
        TestOutOfLimitsVirtualDiskShares(si, options.vmName, 4000, True)

        TestWrongCustomVirtualDiskShares(si, options.vmName)

        TestWrongCustomVirtualDiskLimit(si, options.vmName, 0, False)
        TestWrongCustomVirtualDiskLimit(si, options.vmName, -2, False)
        TestWrongCustomVirtualDiskLimit(si, options.vmName, -1, True)

        TestConfigurationOptions(si, options.vmName)

        TestIncorrectCongestionThreshold(si, 0)
        TestIncorrectCongestionThreshold(si, 4)
        TestIncorrectCongestionThreshold(si, 200)

        TestRemoveDatastore(si)
        TestDeprecatedRemoveDatastore(si)

        print("PASSED.")
    except Exception as e:
        Log("Failed due to: " + str(e))
        print("FAIL.")
    finally:
        cleanup(si, options.vmName)
    print("DONE.")

# Start program
if __name__ == "__main__":
    main()
