#!/usr/bin/python

from __future__ import print_function

import sys, time
import unittest
from pyVmomi import Vim, Vmodl
from pyVim.connect import Connect, Disconnect
from optparse import OptionParser
from pyVim import folder
from pyVim import host

"""
   Host Power Management VIM API Test

   Usage:

   ../py.sh TestPowerSystem.py -H <host name/ip> -u <user name -p <password>

"""

def get_options():
    """
    Supports the command-line arguments listed below
    """

    parser = OptionParser()
    parser.add_option("-H", "--host",
                      default="localhost",
                      help="remote host to connect to")
    parser.add_option("-u", "--user",
                      default="root",
                      help="User name to use when connecting to hostd")
    parser.add_option("-p", "--password",
                      default="",
                      help="Password to use when connecting to hostd")
    (options, _) = parser.parse_args()
    return options


class TestPowerSystem(unittest.TestCase):

   def setOptions(self, options):
      """
      Command line options
      """
      self.options = options

   def setUp(self):
      """
      Setting test suite
      """
      options = get_options()

      self.si = Connect(host=self.options.host,
                        user=self.options.user,
                        version="vim.version.version9",
                        pwd=self.options.password)
      content = self.si.RetrieveContent()
      rootFolder = content.GetRootFolder()
      dataCenter = rootFolder.GetChildEntity()[0]
      hostFolder = dataCenter.hostFolder
      host = hostFolder.childEntity[0]
      self.hostSystem = host.host[0]
      configManager = self.hostSystem.GetConfigManager()
      self.powerSystem = configManager.GetPowerSystem()
      self.hostConfigInfo = self.hostSystem.config

   def tearDown(self):
      """
      Reset test suite
      """
      Disconnect(self.si)

   def runTest(self):
      self.TestCurrentState()
      self.TestSelectPolicy()
      self.TestIncorrectConfigurePolicy()

   def TestCurrentState(self):
      """
      General checks that all the property values are
      populated.
      """

      self.assertTrue(self.powerSystem != None,
                      "power system should not be None")

      self.assertTrue(self.hostConfigInfo != None,
                      "host.config should not be None")

      capability = self.powerSystem.capability
      self.assertTrue(capability != None,
                      "capability should be not None")

      policies = capability.GetAvailablePolicy()
      self.assertTrue(policies != None,
                      "available power policies are not supposed to be None")

      print("number of policies: %d" % len(policies))

      self.assertTrue(self.hostConfigInfo.powerSystemCapability != None,
           "Host.Config.powerSystemCapability should be not None")
      self.assertTrue(self.hostConfigInfo.powerSystemInfo != None,
           "Host.Config.powerSystemInfo should be not None")

      policies2 = self.hostConfigInfo.powerSystemCapability.availablePolicy
      self.assertTrue(len(policies2) == len(policies),
                 "number of policies from PowerSystem and Host.ConfigInfo should be same")

      currentPolicy = self.hostConfigInfo.powerSystemInfo.currentPolicy
      self.assertTrue(currentPolicy != None,
         "Host.Config current policy should not be None")
      self.assertTrue(len(currentPolicy.name) > 0,
         "Host.Config current policy name should not be None")

      currentPolicy = self.powerSystem.info.currentPolicy
      self.assertTrue(currentPolicy != None,
         "Power System current policy should not be None")
      self.assertTrue(len(currentPolicy.name) > 0,
         "PowerSystem current policy should not be None")



   def TestSelectPolicy(self):
      """
      Try to configure every possible power management policy
      """

      print("\nTest select policy")

      policies = self.powerSystem.capability.GetAvailablePolicy()
      self.assertTrue(policies != None,
                      "available power policies are not supposed to be None")

      print("number of policies: %d" % len(policies))

      for p in policies:
        self.assertTrue(len(p.name) > 0)
        self.assertTrue(len(p.shortName) > 0)
        self.assertTrue(len(p.description) > 0)
        self.assertTrue(p.key != None,
                        "power policy key should be not None")

        self.powerSystem.ConfigurePolicy(p.key)
        info = self.powerSystem.info
        self.assertTrue(info != None,
                        "info should be not None")
        newPolicy = info.currentPolicy
        self.assertTrue(newPolicy != None,
                        "current policy can't be None")
        self.assertEqual(newPolicy.key, p.key,
                         "changed policy key should be set")
        self.assertEqual(newPolicy.name, p.name,
                         "changed policy name should be set")
        self.assertEqual(newPolicy.description, p.description,
                         "changed policy description should be set")
        self.assertEqual(newPolicy.key,
           self.hostSystem.config.powerSystemInfo.currentPolicy.key,
           "power system and host config should be in sync")


   def TestIncorrectConfigurePolicy(self):
       """
       Try to configure a policy with incorrect key
       """

       print("\nTest incorrect policy configuration")

       self.assertTrue(self.powerSystem != None,
                       "power system should not be None")

       error = None

       try:
          self.powerSystem.ConfigurePolicy(12345)
       except Exception as e:
          error = e

       self.assertTrue(error != None,
                       "error should not be None")
       self.assertTrue(isinstance(error, Vmodl.Fault.InvalidArgument),
                       "should get invalid argument fault")

def main(argv):
    """
    Test Host Power Management VModl API
    """
    options = get_options()
    test = TestPowerSystem()
    test.setOptions(options)
    suite = unittest.TestSuite()
    suite.addTest(test)
    unittest.TextTestRunner(verbosity=2).run(suite)


# Start program
if __name__ == "__main__":
    main(sys.argv[1:])
