#!/usr/bin/python

# Exercises the date type system code.
# Exercises the DiagnosticManager code.
# Exercises the FirewallSystem code.
# Exercises the ServiceSystem code.
# Exercises the partition manager code.
# Exercises AgentManager.
# Exercises UserManager.
# Exercises AdvancedOptions
# Exercises StorageSystem's iSCSI implementation

# XXX todo:
# o Split these into separate files.
# o Make it easier to track what tests were run and what were not.
# o Add more tests

from __future__ import print_function

import os;
import sys
import time
from pyVmomi import Vim
from pyVmomi import Vmodl
from pyVim import arguments
from pyVim import connect
from pyVim import host
from pyVim import invt
from pyVim import nfclib
from pyVim.connect import SmartConnect, Disconnect
from pyVim.helpers import Log,StopWatch
from pyVim.host import GetRootResourcePool
from pyVim.task import WaitForTask
import time
import atexit

import urllib2          # URL fetching used by DiagnosticManagerTest
import re               # URL patching used by DiagnosticManagerTest


class Phase:
   def __init__(self):
      self.phaseNumber = 0
      self.phaseClock = StopWatch()

   def SetPhase(self, msg):
      Log("Phase " + str(self.phaseNumber) + ": " + msg + " completed")
      self.phaseClock.finish("phase " + str(self.phaseNumber))
      self.phaseClock = StopWatch()
      self.phaseNumber = self.phaseNumber + 1

# Converts a DateTimeInfo structure into a DateTimeConfig
#
# Does not do a deep copy of the initial structure
def DateTimeInfoToConfig(dateTimeInfo):
   dateTimeConfig = Vim.Host.DateTimeConfig()
   dateTimeConfig.SetTimeZone(dateTimeInfo.GetTimeZone().GetKey())
   dateTimeConfig.SetNtpConfig(dateTimeInfo.GetNtpConfig())
   return dateTimeConfig

def IsEmbeddedEsx(testProperties):
   hostType = testProperties["hostType"]
   return hostType == "embeddedEsx"

def IsServiceConsoleEsx(testProperties):
   hostType = testProperties["hostType"]
   return hostType == "esx"

def IsEsx(testProperties):
   return IsEmbeddedEsx(testProperties) or IsServiceConsoleEsx(testProperties)

# Tests for host configuration
class HostConfigTest:
   def __init__(self, args, serviceInstance, hostSystem):
      self._args = args
      self._serviceInstance = serviceInstance
      self._hostSystem = hostSystem
      self._isMockup  = host.IsHostMockup(serviceInstance)

   def Init(self):
      hostSystem = self._hostSystem

      args = self._args
      hostSummary = hostSystem.GetSummary()
      aboutInfo = hostSummary.GetConfig().GetProduct()

      testProperties = {}
      testProperties["hostType"] = aboutInfo.GetProductLineId()
      testProperties["hostName"] = args.GetKeyValue("host")
      testProperties["userName"] = args.GetKeyValue("user")
      testProperties["password"] = args.GetKeyValue("pwd")
      testProperties["script"] = args.GetProgramName()

      self._testProperties = testProperties

   def RunTests(self):
      Log("HostConfigTest: Running host configuration tests")
      self.Init()
      serviceInstance = self._serviceInstance
      hostSystem = self._hostSystem
      testProperties = self._testProperties

      hostConfigManager = hostSystem.GetConfigManager()
      serviceInstanceContent = serviceInstance.RetrieveContent()
      internalServiceInstanceContent = serviceInstance.RetrieveInternalContent()

      dateTimeSystem = hostConfigManager.GetDateTimeSystem()
      if dateTimeSystem:
         dateTimeSystemTest = DateTimeSystemTest(dateTimeSystem, testProperties)
         dateTimeSystemTest.RunTests()
      else:
         Log("Not running DateTimeSystem tests due to a missing managed object.")

      firewallSystem = hostConfigManager.GetFirewallSystem()
      if firewallSystem:
         firewallSystemTest = FirewallSystemTest(firewallSystem, testProperties)
         firewallSystemTest.RunTests()
      else:
         Log("Not running FirewallSystem tests due to a missing managed object.")

      serviceSystem = hostConfigManager.GetServiceSystem()
      if serviceSystem:
         serviceSystemTest = ServiceSystemTest(serviceSystem, testProperties)
         serviceSystemTest.RunTests()
      else:
         Log("Not running ServiceSystem tests due to a missing managed object.")

      storageSystem = hostConfigManager.GetStorageSystem()
      if storageSystem:
         storageSystemTest = StorageSystemTest(storageSystem, testProperties)
         storageSystemTest.RunTests()
      else:
         Log("Not running StorageSystem tests due to a missing managed object.")

      userDirectory = serviceInstanceContent.GetUserDirectory()
      localAccountManager = serviceInstanceContent.GetAccountManager()
      if userDirectory and localAccountManager:
         localAccountManagerTest = LocalAccountManagerTest(userDirectory,
                                                           localAccountManager,
                                                           testProperties)
         localAccountManagerTest.RunTests()
      else:
         Log("Not running LocalAccountManager tests due to a missing managed object.")

      authorizationManager = serviceInstanceContent.GetAuthorizationManager()
      if authorizationManager and localAccountManager:
         permissionTest = PermissionTest(serviceInstanceContent.GetRootFolder(),
                                         authorizationManager,
                                         localAccountManager,
                                         testProperties)
         permissionTest.RunTests()
      else:
         Log("Not running Permission tests due to a missing managed object.")

      optionManager = hostConfigManager.GetAdvancedOption()
      if optionManager:
         optionManagerTest = OptionManagerTest(optionManager, testProperties)
         optionManagerTest.RunTests()
      else:
         Log("Not running OptionManager tests due to a missing managed object.")

      if not self._isMockup:
         diagnosticManager = None
         diagnosticManager = serviceInstanceContent.GetDiagnosticManager()
         if diagnosticManager:
            diagnosticManagerTest = DiagnosticManagerTest(diagnosticManager, testProperties)
            diagnosticManagerTest.RunTests()
         else:
            Log("Not running DiagnosticManager tests due to a missing managed object.")

      agentManager = internalServiceInstanceContent.GetAgentManager()
      nfcService = None
      nfcService = internalServiceInstanceContent.GetNfcService();
      if agentManager:
         agentManagerTest = AgentManagerTest(agentManager, nfcService, testProperties)
         agentManagerTest.RunTests()
      else:
         Log("Not AgentManager tests due to a missing managed object.")


# Tests for DateTimeSystem managed object
class DateTimeSystemTest:
   def __init__(self, dateTimeSystem, testProperties):
      self._dateTimeSystem = dateTimeSystem
      self._testProperties = testProperties

   # Initializes the DateTimeSystemTest before tests are run
   def Init(self):
      dateTimeSystem = self._dateTimeSystem

      # Get initial configuration
      Log("DateTimeSystem: Get current configuration")
      self._initDateTimeInfo = dateTimeSystem.GetDateTimeInfo()

      print(self._initDateTimeInfo)

      self._initDateTimeConfig = DateTimeInfoToConfig(self._initDateTimeInfo)

   def RunTests(self):
      Log("DateTimeSystem: Get current configuration")
      self.Init()

      testProperties = self._testProperties
      dateTimeSystem = self._dateTimeSystem

      dateTimeInfo = self._initDateTimeInfo
      print(dateTimeInfo)

      Log("DateTimeSystem: Get current time")
      initDateTime = dateTimeSystem.QueryDateTime()
      print(initDateTime)

      self.TestUpdateTimezone()

      if not IsEmbeddedEsx(testProperties) :
         self.TestUpdateNtpConfig()
      else:
         Log("DateTimeSystem: Skipping NTP test for embedded ESX.")

      self.TestUpdateDateTime()

   # Returns the timezone of a new timezone.  Selects a timezone
   # that is different than the current one.
   #
   # If no timezone is available, returns None.
   def SelectNewTimeZone(self):
      dateTimeSystem = self._dateTimeSystem
      dateTimeConfig = self._initDateTimeConfig

      if dateTimeConfig == None:
         raise Exception("DateTimeSystemTest -- Must call start before SelectNewTimeZone")

      Log("DateTimeSystem: Get available time zones")
      availTimeZones = dateTimeSystem.QueryAvailableTimeZones()
      #print(availTimeZones)

      currentTzKey = dateTimeConfig.GetTimeZone()

      for tzInfo in availTimeZones:
         if tzInfo.GetKey() != currentTzKey:
            return tzInfo

      return None

   # Test updating time zone configuration
   def TestUpdateTimezone(self):
      dateTimeSystem = self._dateTimeSystem
      initDateTimeConfig = self._initDateTimeConfig

      if initDateTimeConfig == None:
         raise Exception("DateTimeSystemTest -- Must call start before TestUpdateTimezone")

      newTimeZone = self.SelectNewTimeZone()
      if newTimeZone == None:
         Log("DateTimeSystem: Skipping change time zone due to lack of new timezone")
         return

      Log("DateTimeSystem: Changing time zone.")
      dateTimeConfig = Vim.Host.DateTimeConfig()
      dateTimeConfig.SetTimeZone(newTimeZone.GetKey())
      dateTimeSystem.UpdateConfig(dateTimeConfig)

      try:
         dateTimeInfo = dateTimeSystem.GetDateTimeInfo()
         if dateTimeInfo.GetTimeZone().GetKey() != dateTimeConfig.GetTimeZone():
            msg = "Time zone configuration unchanged.  Should be '%s'.  Got '%s'." % \
                  (str(dateTimeConfig.GetTimeZone()), str(dateTimeInfo.GetTimeZone()))
            raise Exception(msg)
      finally:
         # Restore the original configuration
         Log("DateTimeSystem: Restoring original timezone.")
         dateTimeSystem.UpdateConfig(initDateTimeConfig)

   # Test updating NTP configuration
   def TestUpdateNtpConfig(self):
      dateTimeSystem = self._dateTimeSystem
      initDateTimeConfig = self._initDateTimeConfig

      if initDateTimeConfig == None:
         raise Exception("DateTimeSystemTest -- Must call start before TestUpdateTimezone")

      Log("DateTimeSystem: Changing NTP configuration.")
      dateTimeConfig = Vim.Host.DateTimeConfig()
      ntpConfig = Vim.Host.NtpConfig()
      ntpConfig.SetServer(["10.16.65.67"])

      dateTimeConfig.SetNtpConfig(ntpConfig)
      dateTimeSystem.UpdateConfig(dateTimeConfig)

      try:
         dateTimeInfo = dateTimeSystem.GetDateTimeInfo()
         if str(dateTimeInfo.GetNtpConfig()) != str(dateTimeConfig.GetNtpConfig()):
            msg = "NTP configuration unchanged.  Should be '%s'.  Got '%s'." % \
                  (str(dateTimeConfig.GetNtpConfig()), str(dateTimeInfo.GetNtpConfig()))
            raise Exception(msg)
      finally:
         # Restore the original configuration
         Log("DateTimeSystem: Restoring original NTP config.")
         dateTimeSystem.UpdateConfig(initDateTimeConfig)

   def TestUpdateDateTime(self):
      dateTimeSystem = self._dateTimeSystem

      # Tests updating of date time.  Given how time changes, not sure
      # we can do much more than just set it.  Verification expressions
      # cannot be very exact here.
      Log("DateTimeSystem: Update current time")
      dateTime = dateTimeSystem.QueryDateTime()
      dateTimeSystem.UpdateDateTime(dateTime)

      # XXX Do some verification


# Tests for DiagnosticManager managed object
# XXX Need tests for browse() method.
class DiagnosticManagerTest:
   def __init__(self, diagnosticManager, testProperties):
      self._diagnosticManager = diagnosticManager
      self._testProperties = testProperties

   # Initializes the DiagnosticManagerTest before tests are run
   def Init(self):
      diagnosticManager = self._diagnosticManager

      # Get initial configuration
      Log("DiagnosticManager: Get current configuration")

   def RunTests(self):
      self.Init()
      self.TestQueryDescriptions()
      # Too long, plus it's exercised all the time in CAT/SVS.
      # self.TestGenerateLogBundles()

   def TestQueryDescriptions(self):
      Log("DiagnosticManager: Query Descriptions")
      diagnosticManager = self._diagnosticManager

      logDescriptors = diagnosticManager.QueryDescriptions(None)
      print(str(logDescriptors))


   def TestGenerateLogBundles(self):
      Log("DiagnosticManager: Generate Log Bundle")

      diagnosticManager = self._diagnosticManager

      task = diagnosticManager.GenerateLogBundles(True, [])
      Log("DiagnosticManager: Waiting for DiagnosticManager.GenerateLogBundle task to complete.")
      WaitForTask(task, True)

      Log("DiagnosticManager: Task completed successfully.")
      bundleInfos = task.GetInfo().GetResult()
      print(str(bundleInfos))

      Log("DiagnosticManager: Testing validity of bundle URL.")
      if len(bundleInfos) != 1:
         msg = "Invalid number of DiagnosticManager bundle info returned. " \
               "Expecting 1.  Got %d." % bundleInfos.length()
         raise Exception(msg)

      bundleInfo = bundleInfos[0]
      self.TestFetchDiagnosticBundleUrl(bundleInfo.GetUrl())

   def TestFetchDiagnosticBundleUrl(self, url):
      hostname = self._testProperties["hostName"]

      regex = re.compile("\*")
      resolvedUrl = regex.sub(hostname, url)

      # XXX Workaround.  Remove the following two lines
      regex2 = re.compile(":0")
      resolvedUrl = regex2.sub(":80", resolvedUrl)

      Log("DiagnosticManager: Fetching Diagnostic Bundle: %s" % resolvedUrl)
      f = urllib2.urlopen(resolvedUrl)

      # Don't bother actually fetching the data since we don't validate the
      # contents.  Success is determined by the lack of an exception being
      # thrown when the URL is fetched.
      f.close()


# Tests for FirewallSystem managed object
class FirewallSystemTest:
   def __init__(self, firewallSystem, testProperties):
      self._firewallSystem = firewallSystem
      self._testProperties = testProperties

   # Initializes the FirewallSystemTest before tests are run
   def Init(self):
      firewallSystem = self._firewallSystem

      # Get initial configuration
      Log("FirewallSystem: Get current configuration")
      self._initFirewallInfo = firewallSystem.GetFirewallInfo()
      print(str(self._initFirewallInfo))

   def RunTests(self):
      self.Init()
      self.TestUpdateDefaultPolicy()
      self.TestEnableDisableRuleset()

   def TestUpdateDefaultPolicy(self):
      Log("FirewallSystem: Testing Update Default Policy")
      firewallSystem = self._firewallSystem

      firewallInfo = self._initFirewallInfo
      defaultPolicy = firewallInfo.GetDefaultPolicy()

      firewallSystem.UpdateDefaultPolicy(defaultPolicy)

   def TestEnableDisableRuleset(self):
      Log("FirewallSystem: Testing Enable and Disable Ruleset")
      firewallSystem = self._firewallSystem

      firewallInfo = self._initFirewallInfo

      enableCount = 0
      disableCount = 0
      for ruleset in firewallInfo.GetRuleset():
         key = ruleset.GetKey()
         required = ruleset.GetRequired()
         enabled = ruleset.GetEnabled()

         if required and not enabled:
            msg = "FirewallSystem.TestEnableDisableRuleset: Ruleset '%s' is required " \
                  "but not enabled." % key
            raise Exception(msg)

         if required:
            Log("FirewallSystem: Attempting to disable required ruleset %s" % key)

            try:
               firewallSystem.DisableRuleset(key)
            except Exception as e:
               continue

            msg = "FirewallSystem.TestEnableDisableRuleset: Attempt to disable " \
                  "required ruleset '%s' did not result in exception." % key
            raise Exception(msg)

         if enabled and enableCount < 1:
            Log("FirewallSystem: Disabling optional ruleset %s (initially enabled)" % key)
            firewallSystem.DisableRuleset(key)

            Log("FirewallSystem: Enabling optional ruleset %s (initially enabled)" % key)
            firewallSystem.EnableRuleset(key)

            enableCount = enableCount + 1
         elif not enabled and disableCount < 1:
            Log("FirewallSystem: Enabling optional ruleset %s (initially disabled)" % key)
            firewallSystem.EnableRuleset(key)

            Log("FirewallSystem: Disabling optional ruleset %s (initially disabled)" % key)
            firewallSystem.DisableRuleset(key)

            disableCount = disableCount + 1
         else:
            Log("FirewallSystem: Skipping ruleset %s since already tested" % key)

         # XXX Do more validation of ruleset enable/disable

# Tests for ServiceSystem managed object
# XXX Need tests for uninstall
# XXX Need tests for refresh
class ServiceSystemTest:
   def __init__(self, serviceSystem, testProperties):
      self._serviceSystem = serviceSystem
      self._testProperties = testProperties

   # Initializes the ServiceSystemTest before tests are run
   def Init(self):
      serviceSystem = self._serviceSystem

      # Get initial configuration
      Log("ServiceSystem: Get current configuration")
      self._initServiceInfo = serviceSystem.GetServiceInfo()
      print(str(self._initServiceInfo))

   def RunTests(self):
      self.Init()
      self.TestUpdatePolicy()
      self.TestStartRestartStopService()

   def GetScratchService(self):
      serviceSystem = self._serviceSystem
      serviceInfo = self._initServiceInfo

      if serviceInfo == None or serviceInfo.GetService() == None:
         return None

      for service in serviceInfo.GetService():
         # Can't play around with a required service
         if service.GetRequired():
            continue

         # For now, NTP client seems like only service we can play around with
         # and not worry about side effects on the system
         if service.GetKey() == "ntpd":
            return service

      return None

   def TestUpdatePolicy(self):
      Log("ServiceSystem: Testing Update Policy")
      serviceSystem = self._serviceSystem

      service = self.GetScratchService()
      if service == None:
         Log("ServiceSystem: Skipping Update Policy test due to lack of " + \
             "suitable test service.")
         return

      origPolicy = service.GetPolicy()
      key = service.GetKey()
      label = service.GetLabel()

      try:
         for policy in (["on", "automatic", "off"]):
            Log("ServiceSystem: Updating policy of service '%s' to '%s'" % \
                (label, policy))
            serviceSystem.UpdatePolicy(key, policy)
      finally:
         # Restore original policy
         Log("ServiceSystem: Restoring policy of service '%s' to '%s'" % \
             (label, origPolicy))
         serviceSystem.UpdatePolicy(key, origPolicy)

   def TestStartRestartStopService(self):
      Log("ServiceSystem: Testing Start/Restart/Stop of Service")
      serviceSystem = self._serviceSystem

      service = self.GetScratchService()
      if service == None:
         Log("ServiceSystem: Skipping Start/Restart/Stop test due to lack of " + \
             "suitable test service.")
         return

      running = service.GetRunning()
      key = service.GetKey()
      label = service.GetLabel()

      if running:
         Log("ServiceSystem: Stopping running service '%s' in preparation " \
             "for test." % label)
         serviceSystem.Stop(key)

      try:
         Log("ServiceSystem: Test Start Service '%s'" % label)
         serviceSystem.Start(key)

         Log("ServiceSystem: Test Restart Service '%s'" % label)
         serviceSystem.Restart(key)

         Log("ServiceSystem: Test Stop Service '%s'" % label)
         serviceSystem.Stop(key)
      finally:
         if running:
            Log("ServiceSystem: Restoring Service '%s' state by starting " \
                "service." % label)
            serviceSystem.Start(key)


# Tests for StorageSystem managed object
# XXX Need many more tests
class StorageSystemTest:
   def __init__(self, storageSystem, testProperties):
      self._storageSystem = storageSystem
      self._testProperties = testProperties

   # Initializes the StorageSystemTest before tests are run
   def Init(self):
      storageSystem = self._storageSystem

      # Get initial configuration
      Log("StorageSystem: Get current configuration")
      self._initStorageInfo = storageSystem.GetStorageDeviceInfo()

   def RunTests(self):
      self.Init()
      self.TestRetrieveDiskPartitionInfo()
      #self.TestIScsi()

   def TestIscsiAuth(self, hba, targetSet):
      Log("TestIscsiAuth")
      ss = self._storageSystem
      devname = hba.GetDevice()
      if (len(targetSet.staticTargets) > 0):
         obj = targetSet.staticTargets[0]
      elif (len(targetSet.sendTargets) > 0):
         obj = targetSet.sendTargets[0]
      else:
        obj = hba

      ap = hba.GetAuthenticationProperties()
      print("Current chap properties:\n" + repr(ap))

      oldChapSecret = ap.GetChapSecret()
      oldChapName = ap.GetChapName()
      oldChapType = ap.GetChapAuthenticationType()
      oldChapEnabled = ap.GetChapAuthEnabled()
      oldMutualChapSecret = ap.GetMutualChapSecret()
      oldMutualChapName = ap.GetMutualChapName()
      oldMutualChapType = ap.GetMutualChapAuthenticationType()

      ap.SetChapSecret("ifoo")
      ap.SetChapName("ibar")
      ap.SetChapAuthenticationType("chapPreferred")
      ap.SetChapAuthEnabled(True)
      ap.SetMutualChapSecret("mfoo")
      ap.SetMutualChapName("mbar")
      ap.SetMutualChapAuthenticationType("chapRequired")

      # update chap props
      ss.UpdateInternetScsiAuthenticationProperties(devname, targetSet, ap)

      ap.SetChapSecret(oldChapSecret)
      ap.SetChapName(oldChapName)
      ap.SetChapAuthenticationType(oldChapType)
      ap.SetChapAuthEnabled(oldChapEnabled)
      ap.SetMutualChapSecret(oldChapSecret)
      ap.SetMutualChapName(oldMutualChapName)
      ap.SetMutualChapAuthenticationType(oldMutualChapType)

      # restore chap props
      ss.UpdateInternetScsiAuthenticationProperties(devname, targetSet, ap)

   def TestIscsiMisc(self, hba, targetSet):
      Log("TestIscsiMisc")
      ss = self._storageSystem
      devname = hba.GetDevice()

      ip = hba.GetIpProperties()
      Log("Current IP properties:\n" + `ip`)

      ip.SetDelayedAckEnabled(True)
      ip.SetArpRedirectEnabled(True)

      # noop update
      ss.UpdateInternetScsiIPProperties(devname, ip)

   def TestIscsiDigest(self, hba, targetSet):
      Log("TestIscsiDigest")
      ss = self._storageSystem
      devname = hba.GetDevice()
      if (len(targetSet.staticTargets) > 0):
         obj = targetSet.staticTargets[0]
      elif (len(targetSet.sendTargets) > 0):
         obj = targetSet.sendTargets[0]
      else:
        obj = hba

      dp = obj.GetDigestProperties()
      oldHeaderDigestType = dp.GetHeaderDigestType()
      oldDataDigestType = dp.GetDataDigestType()
      print("Current digest properties:\n" + repr(dp))

      dp.SetDataDigestType("digestRequired")
      dp.SetHeaderDigestType("digestPreferred")

      ss.UpdateInternetScsiDigestProperties(devname, targetSet, dp)

      dp.SetDataDigestType(oldDataDigestType)
      dp.SetHeaderDigestType(oldHeaderDigestType)
      ss.UpdateInternetScsiDigestProperties(devname, targetSet, dp)

   def TestIscsiAdvOpts(self, hba, targetSet):
      Log("TestIscsiAdvOpts")
      ss = self._storageSystem
      devname = hba.GetDevice()
      if (len(targetSet.staticTargets) > 0):
         obj = targetSet.staticTargets[0]
      elif (len(targetSet.sendTargets) > 0):
         obj = targetSet.sendTargets[0]
      else:
        obj = hba

      newOpts = []

      supportedOpts = obj.GetSupportedAdvancedOptions()
      opts = obj.GetAdvancedOptions()
      found = -1
      for i in range(len(opts)):
         if opts[i].key == "LogoutTimeout":
            found = i
            break
      if found >= 0:
         oldValue = opts[i].value
         print("Old value = " + repr(oldValue) + " " + repr(type(opts[i].value)))
         opts[i].value = 13
         newOpts.append(opts[i])
         print("Setting to : " + repr(newOpts))
         ss.UpdateInternetScsiAdvancedOptions(devname, targetSet, newOpts)
         raw_input('Hit Enter to continue')
         opts[i].value = oldValue
         ss.UpdateInternetScsiAdvancedOptions(devname, targetSet, newOpts)

   def PrintIScsiAdvOpts(self, opts):
      for opt in opts:
         key = opt.GetKey()
         val = opt.GetValue()
         print("   -- " + key + " : " + val)

   def PrintIScsiSupportedAdvOpts(self, supportedAdvOpts):
      print(supportedAdvOpts)

   def PrintIscsiTarget(self, target, level=1):
      supportedAdvOpts = target.supportedAdvancedOptions
      advOpts = target.advancedOptions
      target.supportedAdvancedOptions = None
      target.advancedOptions = None
      print(target)
      if (level > 2):
         if (supportedAdvOpts):
            self.PrintIScsiAdvOpts(target.supportedAdvancedOptions)
      if (level > 1):
         if (advOpts):
            self.PrintIScsiAdvOpts(advOpts)

   def PrintIscsiHba(self, hba, level=1):
      supportedAdvOpts = hba.supportedAdvancedOptions
      advOpts = hba.advancedOptions
      staticTargets = hba.configuredStaticTarget
      sendTargets = hba.configuredSendTarget

      hba.supportedAdvancedOptions = None
      hba.advancedOptions = None
      hba.configuredStaticTarget = None
      hba.configuredSendTarget = None
      print(hba)
      if (level > 2):
         if (supportedAdvOpts):
            self.PrintIScsiAdvOpts(hba.supportedAdvancedOptions)
      if (level > 1):
         if (advOpts):
            self.PrintIScsiAdvOpts(advOpts)
      for target in sendTargets:
         self.PrintIscsiTarget(target, level)
      for target in staticTargets:
         self.PrintIscsiTarget(target, level)

   def TestConfigureIscsi(self, hba):
      ## test configuration on adapter
      targetSet = Vim.Host.InternetScsiHba.TargetSet()
      self.TestIscsiAuth(hba, targetSet)
      self.TestIscsiDigest(hba, targetSet)
      self.TestIscsiAdvOpts(hba, targetSet)
      self.TestIscsiMisc(hba, targetSet)

      # test configuration on a static target if exists
      staticTargets = hba.configuredStaticTarget
      if len(staticTargets) > 0:
         targetSet = Vim.Host.InternetScsiHba.TargetSet()
         targetSet.staticTargets.append(staticTargets[0])
         self.TestIscsiAuth(hba, targetSet)
         self.TestIscsiDigest(hba, targetSet)
         self.TestIscsiAdvOpts(hba, targetSet)
         self.TestIscsiMisc(hba, targetSet)

      # test configuration on a send target if exists
      # XXX send target config broken?
      sendTargets = hba.configuredSendTarget
      if len(sendTargets) > 0:
         targetSet = Vim.Host.InternetScsiHba.TargetSet()
         targetSet.sendTargets.append(sendTargets[0])
         self.TestIscsiAuth(hba, targetSet)
         self.TestIscsiDigest(hba, targetSet)
         self.TestIscsiAdvOpts(hba, targetSet)
         self.TestIscsiMisc(hba, targetSet)

   def TestIScsi(self):
      Log("StorageSystem: Testing Iscsi Configuration")
      storageSystem = self._storageSystem

      storageInfo = self._initStorageInfo

      hbas = storageInfo.GetHostBusAdapter()
      if len(hbas) < 1:
         Log("StorageSystem: No host bus adapters found ")
         return

      deviceNames = map(lambda x: x.GetDevice(), hbas)
      Log("HBAs found : " +  str(', ').join(deviceNames))

      iscsiHbas = filter(lambda x: isinstance(x, Vim.Host.InternetScsiHba), hbas)
      deviceNames = map(lambda x: x.GetDevice()+":"+x.GetIScsiName(), iscsiHbas)
      Log("ISCSI HBAs found : " +  str(', ').join(deviceNames))

      if len(iscsiHbas) > 0:
         #for hba in iscsiHbas:
         #   self.PrintIscsiHba(hba, 1)
         iHba = iscsiHbas[0]
         self.TestConfigureIscsi(iHba)


   def TestRetrieveDiskPartitionInfo(self):
      Log("StorageSystem: Testing Retrieve Disk Partition Info")
      storageSystem = self._storageSystem

      storageInfo = self._initStorageInfo

      scsiLuns = storageInfo.GetScsiLun()
      scsiDisks = filter(lambda x: isinstance(x, Vim.Host.ScsiDisk), scsiLuns)
      deviceNames = map(lambda x: x.GetDeviceName(), scsiDisks)

      if len(deviceNames) < 1:
         Log("StorageSystem: Skipping Retrieve Disk Partition test due to lack of " + \
             "suitable test storage.")
         return

      Log("StorageSystem: Calling RetrieveDiskPartitionInfo with %d disks: %s" % \
          (len(deviceNames), str(', ').join(deviceNames)))
      diskPartitionInfos = storageSystem.RetrieveDiskPartitionInfo(deviceNames)

      retDeviceNames = map(lambda x: x.GetDeviceName(), diskPartitionInfos)
      Log("StorageSystem: Received DiskPartitionInfo for %d disks: %s" % \
          (len(diskPartitionInfos), str(', ').join(retDeviceNames)))

      # XXX Not sure if the following checks actually work.  Just make them
      # warnings for now.
      if len(diskPartitionInfos) != len(deviceNames):
         msg = "RetrieveDiskPartitionInfo returned different number of return " \
               "values.  Expected %d.  Got %d.\n" % \
               (len(deviceNames), len(diskPartitionInfos))
         Log(msg)
         #raise Exception(msg)

      devicePartitionMap = {}
      for deviceName in deviceNames:
         devicePartitionMap[deviceName] = None

      for diskPartitionInfo in diskPartitionInfos:
         deviceName = diskPartitionInfo.GetDeviceName()

         if not devicePartitionMap.has_key(deviceName):
            msg = "RetrieveDiskPartitionInfo returned partition information " \
                  "for disk '%s' that was not requested." % deviceName
            Log(msg)
            #raise Exception(msg)

         if devicePartitionMap[deviceName] != None:
            msg = "RetrieveDiskPartitionInfo returned duplicate partition " \
                  "information for disk '%s'." % deviceName
            Log(msg)
            #raise Exception(msg)

         devicePartitionMap[deviceName] = diskPartitionInfo


# Tests for LocalAccountManager managed object.  Utilizes the UserDirectory
# to
#
# XXX Need to implement create/remove group tests.
# XXX Need to implement assign/unasssign user to/from group tests.
class LocalAccountManagerTest:
   def __init__(self, userDirectory, localAccountManager, testProperties):
      self._userDirectory = userDirectory
      self._localAccountManager = localAccountManager
      self._testProperties = testProperties

   # Initializes the LocalAccountManagerTest before tests are run
   def Init(self):
      userDirectory = self._userDirectory

      # Should only be invoked on a host that is not attached to AD
      self._allUsers = userDirectory.RetrieveUserGroups(None, "", None, None,
                                                        False, True, False)
      Log("LocalAccountManager: Retrieved %d users." % len(self._allUsers))
      self._allGroups = userDirectory.RetrieveUserGroups(None, "", None, None,
                                                         False, False, True)
      Log("LocalAccountManager: Retrieved %d groups." % len(self._allGroups))


   def RunTests(self):
      self.Init()
      self.TestRetrieveUserGroups()
      self.TestCreateUpdateRemoveUser()

   # Retrieves a UserSearchResult for a user or group verifying that it is the
   # correct spec.
   def RetrieveOneUserGroup(self, principal, isGroup):
      if isGroup:
         principalType = "group"
      else:
         principalType = "user"

      userDirectory = self._userDirectory
      results = userDirectory.RetrieveUserGroups(None, principal, None, None,
                                                 True, not isGroup, isGroup)
      if len(results) != 1:
         msg = "Retrieve did not find existing %s '%s' in search results." % \
               (principalType, principal)
         raise Exception(msg)

      result = results[0]
      if result.GetGroup() != isGroup:
         msg = "Retrieve did not match principal type in search results.  " \
               "Expecting a %s" % principalType
         raise Exception(msg)

      if result.GetPrincipal() != principal:
         msg = "Retrieve did not match existing %s '%s' in search results." \
               "Matched '%s'" % \
               (principalType, results.GetPrincipal(), principal)
         raise Exception(msg)

      return result

   # Make an unused user or group name
   def MakeUnusedPrincipal(self, isGroup):
      if isGroup:
         results = self._allGroups
      else:
         results = self._allUsers

      for i in range(99):
         name = "unitTestUser_%.2d" % i
         match = filter(lambda x: x.GetPrincipal() == name, results)
         if len(match) == 0:
            return name
      return None

   # Make an unused user or group posixId
   # Assumes that the user results are all PosixUserSearchResults
   def MakeUnusedPosixId(self, isGroup):
      if isGroup:
         results = self._allGroups
      else:
         results = self._allUsers

      for i in range(5000,10000):
         match = filter(lambda x: x.GetId() == i, results)
         if len(match) == 0:
            return i
      return None

   # Utility function to convert from LocalAccountManager.AccountSpecification
   # to UserDirectory.UserSearchResults
   #
   # Doesn't actually need to be a method
   def AccountSpecToSearchResult(self, accountSpec, isGroup):
      isPosix = isinstance(accountSpec,
                           Vim.Host.LocalAccountManager.PosixAccountSpecification)
      if isPosix:
         searchResult = Vim.PosixUserSearchResult()
      else:
         searchResult = Vim.UserSearchResult()

      searchResult.SetPrincipal(accountSpec.GetId())
      searchResult.SetFullName(accountSpec.GetDescription())
      searchResult.SetGroup(isGroup)

      if isPosix:
         searchResult.SetId(accountSpec.GetPosixId())
         searchResult.SetShellAccess(accountSpec.GetShellAccess())

      return searchResult

   def TestRetrieveUserGroups(self):
      Log("LocalAccountManager: Testing RetrieveUsersGroups queries")
      userDirectory = self._userDirectory
      allUsers = self._allUsers
      allGroups = self._allGroups

      if len(allUsers) > 1:
         user = allUsers[0]
         userName = user.GetPrincipal()

         Log("LocalAccountManager: Testing exact user match query for '%s'" % userName)
         userResult = self.RetrieveOneUserGroup(userName, False)
         Log("LocalAccountManager: User '%s' matched." % userName)
      else:
         Log("LocalAccountManager: Skipping exact user match query due to lack of users.")

      if len(allGroups) > 1:
         group = allGroups[len(allGroups) - 1]
         groupName = group.GetPrincipal()

         Log("LocalAccountManager: Testing exact group match query for '%s'" % groupName)
         groupResults = self.RetrieveOneUserGroup(groupName, True)
         Log("LocalAccountManager: Group '%s' matched." % groupName)
      else:
         Log("LocalAccountManager: Skipping exact group match query due to lack of groups.")

   def TestCreateUpdateRemoveUser(self):
      userNeedsRemove = False
      try:
         userName = self.MakeUnusedPrincipal(False)
         if userName == None:
            Log("LocalAccountManager: Skipping CreateUpdateRemoveUser test due " + \
                "to lack of unused name.")
            return

         posixId = self.MakeUnusedPosixId(False)
         if posixId == None:
            Log("LocalAccountManager: Skipping CreateUpdateRemoveUser test due " + \
                "to lack of unused UID.")
            return

         # First create the user and verify that it is created
         localAccountManager = self._localAccountManager

         accountSpec = Vim.Host.LocalAccountManager.PosixAccountSpecification()
         accountSpec.SetId(userName)
         accountSpec.SetPassword("nyvra*Znef+yvpr")
         accountSpec.SetDescription("Temporary User for Unit Test")

         accountSpec.SetPosixId(posixId)  # After 5.1 this is ignored.
         accountSpec.SetShellAccess(True) # After 5.1 this is always true.

         Log("LocalAccountManager: Creating user '%s' for create/update/remove test." % \
             userName)
         localAccountManager.CreateUser(accountSpec)

         # Mark that user needs to be removed on errors
         userNeedsRemove = True

         userResult = self.RetrieveOneUserGroup(userName, False)

         # We know that the id will be different, fix for the comparison to pass.
         accountSpec.SetPosixId(userResult.GetId())

         compareResult = self.AccountSpecToSearchResult(accountSpec, False)
         if str(userResult) != str(compareResult):
            Log("Expected: %s\nReceived: %s\n" % (str(compareResult), str(userResult)))
            msg = "User account information for '%s' does not match created user." % \
                  userName
            raise Exception(msg)

         # XXX Should do some tests like see if the it's possible to login with the
         # new user.

         # Second update the user account information
         description = "Changed password for temporary user"
         updateAccountSpec = Vim.Host.LocalAccountManager.PosixAccountSpecification()
         updateAccountSpec.SetId(userName)
         updateAccountSpec.SetDescription(description)
         updateAccountSpec.SetPassword("simple")

         Log("LocalAccountManager: Updating user '%s' for create/update/remove test." % \
             userName)

         Log("LocalAccountManager: Trying weak password")
         try:
            localAccountManager.UpdateUser(updateAccountSpec)
         except Vmodl.Fault.SystemError as e:
            if e.faultMessage[0].key != "com.vmware.vim.host.LocalAccountManager.weakPassword":
               raise Exception("Exception was thrown when setting weak password " \
                               "but had unexpected key.")
            else:
               Log("LocalAccountManager: Got expected exception.")
         else:
            raise Exception("No exception was thrown when setting weak password.")

         Log("LocalAccountManager: Trying strong password")
         updateAccountSpec.SetPassword("NZH3xqTdKy")
         localAccountManager.UpdateUser(updateAccountSpec)

         userResult = self.RetrieveOneUserGroup(userName, False)

         accountSpec.SetDescription(description)
         compareResult = self.AccountSpecToSearchResult(accountSpec, False)
         if str(userResult) != str(compareResult):
            Log("Expected: %s\nReceived: %s\n" % (str(compareResult), str(userResult)))
            msg = "User account information for '%s' does not match updated user." % \
                  userName
            raise Exception(msg)

         # Third remove the user
         Log("LocalAccountManager: Removing user '%s' for create/update/remove test." % \
             userName)
         userNeedsRemove = False
         localAccountManager.RemoveUser(userName)

         Log("LocalAccountManager: Done with create/update/remove test.")
      finally:
         if userNeedsRemove:
            Log("LocalAccountManager: Removing created user '%s' due to errors." % userName)
            localAccountManager.RemoveUser(userName)


class PermissionTest:
   def __init__(self, rootFolder, authorizationManager, localAccountManager,
                testProperties):
      self._rootFolder = rootFolder
      self._authorizationManager = authorizationManager
      self._localAccountManager = localAccountManager
      self._testProperties = testProperties

   def RunTests(self):
      self.TestPermissionCache()
      self.TestViewCache()

   # Test that verifies that hostd updates permission caches for connected
   # sessions when objects move around the inventory. The only movable objects
   # in hostd are the resource pools.
   def TestPermissionCache(self):
      userNeedsRemove = False
      testFailed = True
      rf = self._rootFolder
      am = self._authorizationManager
      lam = self._localAccountManager
      rootSi = connect.GetSi()
      userSi = None
      rootAdmRP = None
      rootRoRP = None
      try:
         userName = "permtestuser"
         password = "nyvra*Znef+yvpr"

         # 1. root creates a ReadOnly user and it logs in.
         accountSpec = Vim.Host.LocalAccountManager.PosixAccountSpecification()
         accountSpec.SetId(userName)
         accountSpec.SetPassword(password)
         accountSpec.SetDescription("Temporary User for Permission Test")

         Log("PermissionTest: Creating user '%s'." % userName)
         lam.CreateUser(accountSpec)
         userNeedsRemove = True

         permission = Vim.AuthorizationManager.Permission()
         permission.group = False
         permission.principal = userName
         permission.propagate = True
         permission.roleId = -2 # ROLE_READONLY_ID
         permissionArr = [permission]

         Log("PermissionTest: Giving ReadOnly role to user '%s' "
             "on the root folder." % userName)
         am.SetEntityPermissions(rf, permissionArr)

         Log("PermissionTest: Loggin in user '%s'." % userName)
         # userSi is from the permtestuser's session.
         userSi = SmartConnect(host=self._testProperties["hostName"],
                                     user=userName,
                                     pwd=password)

         # 2. root creates two resource pools - pool-adm and pool-ro - and
         #    gives the user a propagating Admin role over pool-adm.
         spec = Vim.ResourceConfigSpec()
         spec.entity = None
         resAlloc = Vim.ResourceAllocationInfo()
         resAlloc.reservation = 0
         resAlloc.expandableReservation = True
         resAlloc.limit = -1
         resAlloc.shares = Vim.SharesInfo()
         resAlloc.shares.level = Vim.SharesInfo.Level.normal
         resAlloc.shares.shares = 1984
         spec.cpuAllocation = resAlloc
         spec.memoryAllocation = resAlloc

         rootRP = GetRootResourcePool(rootSi)
         Log("PermissionTest: root creating pool-adm.")
         rootAdmRP = rootRP.CreateResourcePool("pool-adm", spec)
         Log("PermissionTest: root creating pool-ro.")
         rootRoRP = rootRP.CreateResourcePool("pool-ro", spec)

         permission.roleId = -1 # ROLE_ADMIN_ID
         Log("PermissionTest: Giving Admin role to user '%s' "
             "on pool-adm." % userName)
         am.SetEntityPermissions(rootAdmRP, permissionArr)

         # 3. The user creates pool-l2, pool-l3 chain under pool-adm.
         userAdmRP = invt.findResourcePools('', "pool-adm", userSi)[0][0]
         Log("PermissionTest: user creating pool-l2 under pool-adm.")
         userPoolL2 = userAdmRP.CreateResourcePool("pool-l2", spec)
         Log("PermissionTest: user creating pool-l3 under pool-l2.")
         userPoolL3 = userPoolL2.CreateResourcePool("pool-l3", spec)
         # IMPORTANT: Call a method to populate the auth cache.
         userPoolL3.DestroyChildren()

         # 4. root moves pool-l2 under pool-ro.
         rootPoolL2 = invt.findResourcePools('', "pool-l2", rootSi)[0][0]
         Log("PermissionTest: root moving pool-l2 under pool-ro.")
         rootRoRP.MoveInto([rootPoolL2])

         # 5. The user tries to create pool-l4 under under pool-l3 and fails.
         #    pool-l3 is used in order to check that NotifyEntityMove
         #    was called for the children of pool-l2 too.
         Log("PermissionTest: user creating pool-l4 under pool-l3.")
         try:
            userPoolL4 = userPoolL3.CreateResourcePool("pool-l4", spec)
         except Vmodl.Fault.SecurityError as e:
            Log("PermissionTest: Caught expected SecurityError.")
         else:
            raise Exception("PermissionTest: No exception was thrown when "
                            "creating a pool without rights. hostd is not "
                            "calling NotifyEntityMove properly.")

         # 6. root creates pool-l4 under under pool-l3 and moves pool-l2.
         #    back under pool-adm
         Log("PermissionTest: root creating pool-l4 under pool-l3.")
         rootPoolL3 = invt.findResourcePools('', "pool-l3", rootSi)[0][0]
         rootPoolL4 = rootPoolL3.CreateResourcePool("pool-l4", spec)
         Log("PermissionTest: root moving pool-l2 back under pool-adm.")
         rootAdmRP.MoveInto([rootPoolL2])

         # 7. The user deletes pool-l2.
         Log("PermissionTest: user removing pool-l2.")
         userPoolL2.Destroy()

         # 8. Cleanup
         testFailed = False
         Log("PermissionTest: Test passed.")

      finally:
         # Log out permtestuser and restore the global si
         Disconnect(userSi)
         connect.SetSi(rootSi)
         if testFailed:
            Log("PermissionTest: Test failed.")
         if rootAdmRP:
            Log("PermissionTest: Removing pool-adm.")
            rootAdmRP.Destroy()
         if rootRoRP:
            Log("PermissionTest: Removing pool-ro.")
            rootRoRP.Destroy()
         if userNeedsRemove:
            Log("PermissionTest: Removing user '%s'." % userName)
            lam.RemoveUser(userName)

   # Test that verifies that hostd updates System.View caches for connected
   # sessions when objects move around the inventory. The only movable objects
   # in hostd are the resource pools.
   def TestViewCache(self):
      userNeedsRemove = False
      testFailed = True
      rf = self._rootFolder
      am = self._authorizationManager
      lam = self._localAccountManager
      rootSi = connect.GetSi()
      userSi = None
      rootA1RP = None
      rootB1RP = None
      try:
         userName = "permtestuser"
         password = "nyvra*Znef+yvpr"

         # 1. root creates two chains of resource pools from the root pool.
         spec = Vim.ResourceConfigSpec()
         spec.entity = None
         resAlloc = Vim.ResourceAllocationInfo()
         resAlloc.reservation = 0
         resAlloc.expandableReservation = True
         resAlloc.limit = -1
         resAlloc.shares = Vim.SharesInfo()
         resAlloc.shares.level = Vim.SharesInfo.Level.normal
         resAlloc.shares.shares = 1984
         spec.cpuAllocation = resAlloc
         spec.memoryAllocation = resAlloc

         rootRP = GetRootResourcePool(rootSi)
         Log("PermissionTest: root creating pool-a1 under root.")
         rootA1RP = rootRP.CreateResourcePool("pool-a1", spec)
         Log("PermissionTest: root creating pool-a2 under pool-a1.")
         rootA2RP = rootA1RP.CreateResourcePool("pool-a2", spec)
         Log("PermissionTest: root creating pool-b1 under root.")
         rootB1RP = rootRP.CreateResourcePool("pool-b1", spec)
         Log("PermissionTest: root creating pool-b2 under pool-b1.")
         rootB2RP = rootB1RP.CreateResourcePool("pool-b2", spec)

         # 2. root creates a ReadOnly user and gives it permission on pool-a2.
         #    The user logs in, but can't find pool-b1.
         accountSpec = Vim.Host.LocalAccountManager.PosixAccountSpecification()
         accountSpec.SetId(userName)
         accountSpec.SetPassword(password)
         accountSpec.SetDescription("Temporary User for Permission Test")

         Log("PermissionTest: Creating user '%s'." % userName)
         lam.CreateUser(accountSpec)
         userNeedsRemove = True

         permission = Vim.AuthorizationManager.Permission()
         permission.group = False
         permission.principal = userName
         permission.propagate = True
         permission.roleId = -2 # ROLE_READONLY_ID
         permissionArr = [permission]

         Log("PermissionTest: Giving ReadOnly role to user '%s' "
             "on pool-a2." % userName)
         am.SetEntityPermissions(rootA2RP, permissionArr)

         Log("PermissionTest: Loggin in user '%s'." % userName)
         # userSi is from the permtestuser's session.
         userSi = SmartConnect(host=self._testProperties["hostName"],
                                     user=userName,
                                     pwd=password)

         if len(invt.findResourcePools('', "pool-b1", userSi)) == 0:
            Log("PermissionTest: User expectedly cannot find pool-b1.")
         else:
            raise Exception("PermissionTest: User can find pool-b1 "
                            "without rights.")

         # 3. root gives the user permission on pool-b2.
         #    He should now have System.View on pool-b1.
         Log("PermissionTest: Giving ReadOnly role to user '%s' "
             "on pool-b2." % userName)
         am.SetEntityPermissions(rootB2RP, permissionArr)

         if len(invt.findResourcePools('', "pool-b1", userSi)) == 1:
            Log("PermissionTest: User can find pool-b1.")
         else:
            raise Exception("PermissionTest: User cannot find pool-b1 "
                            "with rights.")

         # 3. root moves pool-b2 under pool-a1.
         #    The user should stop seeing pool-b1.
         Log("PermissionTest: root moving pool-b2 under pool-a1.")
         rootA1RP.MoveInto([rootB2RP])

         if len(invt.findResourcePools('', "pool-b1", userSi)) == 0:
            Log("PermissionTest: User expectedly cannot find pool-b1.")
         else:
            raise Exception("PermissionTest: User can find pool-b1 "
                            "without rights.")

         # 4. root creates pool-b3 under pool-b1, gives permission to the user
         #    on the new pool and deleted the pool.
         #    The user should not see pool-b1
         Log("PermissionTest: root creating pool-b3 under pool-b1.")
         rootB3RP = rootB1RP.CreateResourcePool("pool-b3", spec)

         Log("PermissionTest: Giving ReadOnly role to user '%s' "
             "on pool-b3." % userName)
         am.SetEntityPermissions(rootB3RP, permissionArr)

         Log("PermissionTest: Removing pool-b3.")
         rootB3RP.Destroy()

         if len(invt.findResourcePools('', "pool-b1", userSi)) == 0:
            Log("PermissionTest: User expectedly cannot find pool-b1.")
         else:
            raise Exception("PermissionTest: User can find pool-b1 "
                            "without rights.")

         # 5. Cleanup
         testFailed = False
         Log("PermissionTest: Test passed.")
         return

      finally:
         # Log out permtestuser and restore the global si
         Disconnect(userSi)
         connect.SetSi(rootSi)
         if testFailed:
            Log("PermissionTest: Test failed.")
         if rootA1RP:
            Log("PermissionTest: Removing pool-a1.")
            rootA1RP.Destroy()
         if rootB1RP:
            Log("PermissionTest: Removing pool-b1.")
            rootB1RP.Destroy()
         if userNeedsRemove:
            Log("PermissionTest: Removing user '%s'." % userName)
            lam.RemoveUser(userName)

class OptionManagerTest:
   def __init__(self, optionManager, testProperties):
      self._optionManager = optionManager
      self._testProperties = testProperties

   # Initializes the OptionManagerTest before tests are run
   def Init(self):
      optionManager = self._optionManager

      self._optionMap = {}

      Log("OptionManagerTest: Fetching all supported options")
      optionDefs = optionManager.GetSupportedOption()
      for optionDef in optionDefs:
         key = optionDef.GetKey()
         self._optionMap[key] = { "DEF": optionDefs,
                                  "VALUE": None }

      Log("OptionManagerTest: Fetching all settings")
      settings = optionManager.GetSetting()
      for setting in settings:
         key = setting.GetKey()
         if not self._optionMap[key]:
            msg = "OptionManager: OptionValue '%s' not defined by OptionDef" % key
            raise Exception(msg)
         self._optionMap[key]["VALUE"] = setting.GetValue()

      Log("OptionManagerTest: Loaded %d options" % len(self._optionMap.keys()))

   def RunTests(self):
      self.Init()

      print(self.GetOptionTreeString())

   def GetOptionTreeString(self):
      s = []
      optionKeys = self._optionMap.keys()
      optionKeys.sort()
      for optionKey in optionKeys:
         optionValue = self._optionMap[optionKey]["VALUE"]
         s.append("%s = %s" % (optionKey, optionValue))

      return str("\n").join(s);

   def GetOption(self, key):
      return self._optionMap[key]

   def GetOptionValue(self, key):
      if self._optionMap[key] != None:
         return self._optionMap[key]["VALUE"]
      return None;

   def SetOptionValue(self, key, value):
      pass


   def TestSyslogConfig(self):

      # Test Syslog.Remote.Enabled
      # Test Syslog.Remote.Hostname
      # Test Syslog.Remote.Port

      # Test Syslog.Local.Enabled
      # Test Syslog.Local.DatastorePath

      pass


# Test requires that SSL be disabled because the Python implementation of
# SSL doesn't support the cryptography used by Authd.
class AgentManagerTest:
   def __init__(self, agentManager, nfcService, testProperties):
      self._agentManager = agentManager
      self._nfcService = nfcService
      self._testProperties = testProperties

   # Initializes the AgentManagerTest before tests are run
   def Init(self):
      agentManager = self._agentManager

   def RunTests(self):
      self.Init()

      self.TestUpgrade()

   def ConnectNfcClient(self):
      nfcService = self._nfcService
      testProperties = self._testProperties

      Log("AgentManagerTest: Acquiring ticket for NFC use");
      hostServiceTicket = nfcService.SystemManagement(None)

      Log("AgentManagerTest: Ticket received. %s" % str(hostServiceTicket));

      if hostServiceTicket.GetHost() == None:
         hostServiceTicket.SetHost(testProperties["hostName"])

      if hostServiceTicket.GetPort() == None:
         hostServiceTicket.SetPort(902)

      Log("AgentManagerTest: Creating and connecting NFC client");
      nfcClient = nfclib.NfcClient(hostServiceTicket)
      nfcClient.Connect()

      return nfcClient


   def TestUpgrade(self):
      nfcClient = None
      try:
         nfcClient = self.ConnectNfcClient()
      except Exception as e:
         Log("AgentManagerTest: Failed to connect NfcClient.  %s" % str(e))
         Log("AgentManagerTest: Skipping Agent Manager upgrade test.")
         return

      agentManager = self._agentManager
      testProperties = self._testProperties
      Log("AgentManagerTest: Testing upgrade");

      tempPath = agentManager.PrepareToUpgrade()
      Log("AgentManagerTest: Using temporary upgrade path '%s'" % tempPath);

      dataPath = os.path.dirname(testProperties["script"]) + "/data"
      executable = "spew.sh"
      signature = "spew.sig"
      fileset = [(dataPath + "/spew.sh", tempPath + "/" + executable),
                 (dataPath + "/spew.sig", tempPath + "/" + signature)]

      Log("AgentManagerTest: Putting files %s and %s on host using NFC" %
          (tempPath + "/" + executable, tempPath + "/" + signature))
      if not nfcClient.PutFiles(fileset):
         raise Exception("Failed to put files on host.")

      Log("AgentManagerTest: Invoking upgrade")
      agentManager.Upgrade(executable, signature)

      Log("AgentManagerTest: Successfully tested upgrade.");



def main():
   supportedArgs = [ (["h:", "host="], "localhost", "Host name", "host"),
                     (["u:", "user="], "root", "User name", "user"),
                     (["p:", "pwd="], "", "Password", "pwd"),
                     (["i:", "numiter="], "1", "Number of iterations", "iter") ]

   supportedToggles = [ (["usage", "help"], False, "Show usage information", "usage"),
                        (["cleanup", "c"], True, "Try to cleanup test vms from previous runs", "cleanup")]

   args = arguments.Arguments(sys.argv, supportedArgs, supportedToggles)
   if args.GetKeyValue("usage") == True:
      args.Usage()
      sys.exit(0)

   # Connect
   si = SmartConnect(host=args.GetKeyValue("host"),
                     user=args.GetKeyValue("user"),
                     pwd=args.GetKeyValue("pwd"))
   atexit.register(Disconnect, si)

   Log("Connected to host " + args.GetKeyValue("host"))

   # Process command line
   numiter = int(args.GetKeyValue("iter"))
   doCleanup = args.GetKeyValue("cleanup")
   status = "PASS"

   resultsArray = []

   for i in range(numiter):
      bigClock = StopWatch()
      try:
         try:
            ph = Phase()

            # Get the HostSystem
            hostSystem = host.GetHostSystem(si)
            hostConfigTest = HostConfigTest(args, si, hostSystem)
            hostConfigTest.RunTests()
            ph.SetPhase("Host Config Tests")

            status = "PASS"

         finally:
            bigClock.finish("iteration " + str(i))

      # While debugging, uncomment the line below to see backtraces
      # when an exception occurs.
#      except int as e:
      except Exception as e:
         Log("Caught exception : " + str(e))
         status = "FAIL"

      Log("TEST RUN COMPLETE: " + status)
      resultsArray.append(status)

   Log("Results for each iteration: ")
   for i in range(len(resultsArray)):
      Log("Iteration " + str(i) + ": " + resultsArray[i])

# Start program
if __name__ == "__main__":
    main()
