#!/usr/bin/python

import sys
import getopt
from pyVmomi import Vim
from pyVim.connect import Connect, Disconnect
from pyVim.task import WaitForTask
from pyVim import vm, folder, invt
from pyVim import vmconfig
from pyVim import arguments
from pyVim.helpers import Log,StopWatch
import atexit

## Cleanup existing VMs with given name
def CleanupVm(vmname):
    vm1 = folder.Find(vmname)
    if vm1 != None:
        Log("Cleaning up VM " + vmname)
        vm.Destroy(vm1)


## Test PCI passthrough device support
def TestPCIPassthroughDevice(vm1, i):
    failed = False
    # Test adding PCI passthrough device to VM
    Log("Adding PCI passthrough device to VM")
    cspec = Vim.Vm.ConfigSpec()
    curMem = vm1.GetConfig().GetHardware().GetMemoryMB()
    memAlloc = Vim.ResourceAllocationInfo()
    memAlloc.SetReservation(long(curMem))
    memAlloc.SetLimit(long(curMem))
    cspec = vmconfig.AddPCIPassthrough(cspec)
    task = vm.Reconfigure(vm1, cspec)

    Log("Checking for presence of PCI Passthrough device in VM")
    devices = vmconfig.CheckDevice(vm1.GetConfig(),
                                   Vim.Vm.Device.VirtualPCIPassthrough)
    if len(devices) != i:
        raise Exception("VM has " + str(len(devices)) + \
                        "PCI passthrough devices. Expected " + str(i))
    Log("Verified PCI passthrough device was added to the VM")


def main():
   supportedArgs = [ (["h:", "host="], "localhost", "Host name", "host"),
                     (["u:", "user="], "root", "User name", "user"),
                     (["p:", "pwd="], "ca$hc0w", "Password", "pwd"),
                     (["v:", "vmname="], "Hw7ReconfigTest", "Name of the virtual machine", "vmname"),
                     (["i:", "numiter="], "1", "Number of iterations", "iter") ]

   supportedToggles = [ (["usage", "help"], False, "Show usage information", "usage"),
                        (["runall", "r"], True, "Run all the tests", "runall"),
                        (["nodelete"], False, "Dont delete vm on completion", "nodelete") ]

   args = arguments.Arguments(sys.argv, supportedArgs, supportedToggles)
   if args.GetKeyValue("usage") == True:
      args.Usage()
      sys.exit(0)

   # Connect
   si = Connect(host=args.GetKeyValue("host"),
                user=args.GetKeyValue("user"),
                pwd=args.GetKeyValue("pwd"),
                version="vim.version.version9")
   atexit.register(Disconnect, si)


   # Process command line
   vmname = args.GetKeyValue("vmname")
   numiter = int(args.GetKeyValue("iter"))
   runall = args.GetKeyValue("runall")
   noDelete = args.GetKeyValue("nodelete")
   status = "PASS"

   for i in range(numiter):
       bigClock = StopWatch()
       vm1 = None
       try:

	   ## Determine max number of PCI passthrough devices supported on host
	   envBrowser = invt.GetEnv()
	   max = 0
	   cfgOption = envBrowser.QueryConfigOption(None, None)
	   for opt in cfgOption.GetHardwareOptions().GetVirtualDeviceOption():
	       if isinstance(opt, Vim.Vm.Device.VirtualPCIControllerOption):
		  max = opt.GetNumPCIPassthroughDevices().GetMax()
		  break
	   if max == 0:
	      raise Exception("No PCI Passthrough devices supported on the host!")
	   Log("Maximum number of PCI Passthrough devices supported: " + str(max))

           ## Cleanup old VMs
           posVmName = vmname + "_Pos_" + str(i)
           CleanupVm(posVmName)

           ## Positive tests on a hwVersion 7 VM
           Log("Creating Hw7 VM..")
           vm1 = vm.CreateQuickDummy(posVmName, 0, 1,
                                     vmxVersion = "vmx-07")
	   for j in range(1, max + 1):
	      TestPCIPassthroughDevice(vm1, j)

           Log("Destroying VM")
           vm.Destroy(vm1)
           Log("Tests completed.")
           bigClock.finish("iteration " + str(i))
       except Exception as e:
           status = "FAIL"
           Log("Caught exception : " + str(e))
   Log("TEST RUN COMPLETE: " + status)


# Start program
if __name__ == "__main__":
    main()
